##############################################################################
# resolutions.py - managing collections of resolutions
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#        
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License at http://www.gnu.org/licenses/gpl.txt
# By using, editing and/or distributing this software you agree to
# the terms and conditions of this license.

# define sorted() for Python 2.3 and below
try: sorted(())
except:
    def sorted(obj):
        lnew = type(obj)(obj)
        lnew.sort()
        return lnew

class Resolution:
    '''a single resolution with a width, height, and a sort weight.
    Resolutions can be initialised from a string or a list and can be sorted.'''

    physical = None
    virtual = None
    weight = 0      # how 'important' this resolution is for sorting
	
    def __init__(self, val=None, weight=None):
        self.set(val, weight)

    def set(self, val, weight=None):
        '''set resolution. Value can be either None, a string in the form
        "WIDTHxHEIGHT", "WIDTHxHEIGHT @VWIDTHxVHEIGHT", another Resolution
        instance, a list [width,height] or a list
        [[width,height],[vwidth,vheight].
        When weight is not None, it will be set.'''
        self.physical = self.virtual = None
        if weight != None: self.weight = weight
        if isinstance(val, Resolution):
            if val.physical: self.physical = val.physical[0], val.physical[1]
            if val.virtual: self.virtual = val.virtual[0], val.virtual[1]
            self.weight = val.weight
        elif val == None:
            pass
        elif type(val) == str:
            parts = val.split('@')
            self.physical = map(int, parts[0].split('x',1))
            if len(parts)>1:
                self.virtual = map(int, parts[1].split('x',1))
            elif len(parts)>2:
                raise TypeError('need zero or one virtual resolutions (@): '+val)
        elif type(val) in [list,tuple]:
            if type(val[0]) in [list,tuple]:
                self.physical,self.virtual = val
            else:
                self.physical = val
        else:
            raise TypeError('invalid resolution: '+str(val))
        # validate
        if len(self.physical)!=2:
            raise TypeError('invalid physical resolution in: '+str(val))
        if self.virtual and len(self.virtual)!=2:
            raise TypeError('invalid virtual resolution in: '+str(val))
        # make sure we have fixed type to avoid comparison trouble
        self.physical = tuple(self.physical)
        if self.virtual: self.virtual = tuple(self.virtual)

    def size(self):
        return self.physical

    def __str__(self):
        s = '%dx%d'%(self.physical[0],self.physical[1])
        if self.virtual and self.physical != self.virtual:
            s += ' @%dx%d'%(self.virtual[0],self.virtual[1])
        return s

    def __eq__(self, other):
        if not isinstance(other, Resolution):
            try: other = Resolution(other)
            except: return False
        virt1,virt2 = self.virtual, other.virtual
        if not virt1: virt1 = self.physical
        if not virt2: virt2 = other.physical
        return (self.physical == other.physical) and (virt1 == virt2)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __cmp__(self, other):
        # sort order: by weight first, then physical size, then virtual size
        if not isinstance(other, Resolution):
            other = Resolution(other)
        diff = self.weight - other.weight
        if diff: return diff
        diff = self.physical[0]*self.physical[1] - other.physical[0]*other.physical[1]
        if diff: return diff
        virt1,virt2 = self.virtual, other.virtual
        if not virt1: virt1 = self.physical
        if not virt2: virt2 = other.physical
        diff = virt1[0]*virt1[1] - virt2[0]*virt2[1]
        if diff: return diff
        # equal
        return 0

    def __hash__(self):
        return hash(self.__str__())


class ResolutionList(list):
    '''a list of resolutions'''

    def __init__(self, val=None):
        self.set(val)

    def set(self, val):
        self[:] = []
        if val == None: return
        if type(val) == str:
            val = val.split(',')
        for v in val:
            self.append(Resolution(v))

    def __str__(self):
        return ', '.join(map(str, self))

    def __eq__(self, other):
        if not isinstance(other, ResolutionList):
            other = ResolutionList(other)
        return list(self) == list(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __cmp__(self, other):
        if not isinstance(other, ResolutionList):
            other = ResolutionList(other)
        return cmp(list(self), list(other))


class ResolutionSelection(dict):
    '''a hash containing one resolution for each display'''

    def __init__(self, sress=None, displays=None):
        '''create new ResolutionSelection, optionally from a string of
        resolutions in the form "800x600, 240x340", one resolution for
        each display. Alternatively if sress contains only one
        resolution, this is used for each display.'''
        self.set(sress, displays)

    def set(self, sress, displays=None):
        self.clear()
        if not sress: return
        if isinstance(sress, ResolutionSelection):
            # copy from other object
            for i in sress: self.append(i)
        elif type(displays) in [list,tuple]:
            if type(sress)==str: 
                sress = sress.split(',')
            if type(sress) not in [list,tuple]:
                sress = [sress]
            if len(sress) == 1:
                sress = sress * len(displays)
            if len(sress) != len(displays):
                raise ValueError('number of resolutions must be equal to number of displays')
            for i,sr in enumerate(sress):
                self[displays[i]] = Resolution(sr)
        else:
            raise ValueError('unrecognised resolution selection')

    def __str__(self):
        s = []
        for disp,r in self.iteritems():
            s.append('%s: %s'%(disp, r))
        return ', '.join(s)


class ResolutionCollection(dict):
    '''a hash containing a list of resolutions for each display'''

    def __str__(self):
        s = []
        for disp,rl in self.iteritems():
            s.append('%s: %s'%(disp, rl))
        return '\n'.join(s)

    def common(self):
        '''return ResolutionList with only resolutions common to all displays.
        The weight of each resolution is the sum of that of each display.'''
        if len(self) == 0: return ResolutionList()
        common = None
        for disp,rl in self.iteritems():
            rl = ResolutionList(rl)
            if not common:
                common = rl
                continue
            delitems = ResolutionList()
            for c in common:
                if c in rl:
                    c.weight += rl[rl.index(c)].weight
                else:
                    delitems.append(c)
            for d in delitems: common.remove(d)
        return common

    def sort(self):
        '''sort all contained resolution lists'''
        for disp,rl in self.iteritems():
            rl.sort()

    def reverse(self):
        '''reverse all contained resolution lists'''
        for disp,rl in self.iteritems():
            rl.reverse()

    def select(self, choose=lambda disp,rl: sorted(rl)[-1]):
        '''return a dict containing the topmost resolution for each display.
        This is useful to get an actual resolution for each display after having
        sorted each contained ResolutionList with the best resolution at the
        end. Or if another selection criterion is wanted, it can be specified.'''
        r = ResolutionSelection()
        for disp,rl in self.iteritems():
            r[disp] = choose(disp, rl)
        return r


# some testing
if __name__ == '__main__':
    r = Resolution('800x600')
    if r != Resolution('800x600'):
        print 'ERROR: resolution string init, resolution compare: %s'%r
    if r != [800, 600]:
        print 'ERROR: resolution string init, list compare: %s'%r
    if r != '   800    x 600  ':
        print 'ERROR: resolution string init, string compare: %s'%r

    if not Resolution([1024, 768]) > r:
        print 'ERROR: resolution compare'
    if Resolution([640,480], 1) <= r:
        print 'ERROR: sort order 1 vs 0'
    if Resolution([640,480], -5) >= r:
        print 'ERROR: sort order -5 vs 0'

    rl = ResolutionList(' 640x480,    1024   x   768 , 800x600, 200x300  ')
    rl2 = ResolutionList(['640x480', '1024x768', '800x600', '200x300'])
    rl3 = ResolutionList([[640,480], [1024,768], [800,600], [200,300]])
    if rl != rl2:
        print 'ERROR: resolutionlist string vs. list of strings'
    if rl != rl3:
        print 'ERROR: resolutionlist string vs. list of lists'
    if '800x600' not in rl:
        print 'ERROR: cannot find string resolution in list'
    if [800,600] not in rl:
        print 'ERROR: cannot find list resolution in list'

    rl.sort()
    if rl != '200x300, 640x480, 800x600, 1024x768':
        print 'ERROR: resolution list sort: %s'%rl

    rc = ResolutionCollection()
    rl1 = ResolutionList('640x480, 200x300, 600x800, 1024x800, 400x800, 400x300')
    rl2 = ResolutionList('800x600, 1000x800, 200x300, 400x300, 500x300')
    rc['display1'] = rl1
    rc['display2'] = rl2

    if rc.common() != '200x300, 400x300':
        print 'ERROR: resolution collection common'
    if rc.select() != {'display1': Resolution('1024x800'), 'display2': Resolution('1000x800')}:
        print 'ERROR: resolution collection select'

    rcsorted = ResolutionCollection()
    rcsorted['display1'] = sorted(rl1)
    rcsorted['display2'] = sorted(rl2)
    if rc == rcsorted:
        print 'ERROR: resolution collections should not be equal before sorting'
    rc.sort()
    if rc != rcsorted:
        print 'ERROR: inplace sorting failed'

    print 'all tests done.'

# vim:ts=4:sw=4:expandtab:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Python-XRandR provides a high level API for the XRandR extension of the
# X.org server. XRandR allows to configure resolution, refresh rate, rotation 
# of the screen and multiple outputs of graphics cards.
#
# This module allows to get information for gtk.gdk.Screen objects
#
# In many aspects it follows the design of the xrand tool written by
# Keith Packard.
#
# Copyright 2007 © Sebastian Heinlein <sebastian.heinlein@web.de>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

import pygtk
pygtk.require("2.0")
import gtk.gdk

import xrandr

#FIXME: Perhaps using gdk_x11_screen_get_xscreen would be more elegant

def get_default_screen_config():
    dpy = gtk.gdk.display_get_default()
    screen = dpy.get_default_screen()
    return get_screen_config(screen)

def get_screen_config(screen):
    """Returns the XRandR screen config instance for the given gtk.gdk.Screen"""
    dpy = screen.get_display()
    dpy_url = dpy.get_name()
    count = screen.get_number()
    return xrandr.get_screen_of_display(dpy_url, count)

# vim:ts=4:sw=4:et



#!/usr/bin/env python
###############################################################################
# disper.py - main disper
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#        
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License at http://www.gnu.org/licenses/gpl.txt
# By using, editing and/or distributing this software you agree to
# the terms and conditions of this license.

import os
import sys
import logging
import optparse
import shlex

from switcher import Switcher, Resolution, ResolutionSelection
from plugins import Plugins
import build

# make sure to handle SystemExit when using this class in a 3rd-party program

class Disper:

    # static information
    name = 'disper'
    version = '0.3.1'
    prefix = build.prefix
    prefix_share = build.prefix_share

    # option parsing
    argv = []
    parser = None           # option parser object
    options = None          # parsed options
    args = None             # parsed arguments

    # real work
    _switcher = None        # switcher object, see Disper.switcher()
    plugins = None          # plugins object
    log = None
    conffile = None         # last configuration file read

    def __init__(self):
        self.log = logging.getLogger('disper')
        self.log.setLevel(logging.WARNING)
        self._options_init()
        self.plugins = Plugins(self)
        #self.plugins.call('init') # can't really do here since list of plugins isn't read yet
        # add default options
        # TODO do initial parsing too so errors can be traced to config
        self.options_append(self.config_read_default())

    def _options_init(self):
        '''initialize default command-line options'''
        usage = "usage: %prog [options] (-l|-s|-c|-e|-p|-i)"
        version = ' '.join(map(str, [self.name, self.version]))
        self.parser = optparse.OptionParser(usage=usage, version=version)

        self.add_option('-v', '--verbose', action='store_const', dest='debug', const=logging.INFO,
            help='show what\'s happening')
        self.add_option('-q', '--quiet', action='store_const', dest='debug', const=logging.ERROR,
            help='be quiet and only show errors')
        self.add_option('-r', '--resolution', dest='resolution',  
            help='set resolution, e.g. "800x600", or "auto" to detect the display\'s preferred '+
                 'resolution, "max" to use the maximum resolution advertised, or "off" to disable '+
                 'the display entirely. For extend it is possible to enter a single resolution for '+
                 'all displays or a comma-separated list of resolutions (one for each display). '+
                 'Beware that many displays advertise resolutions they can not fully show, '+
                 'so "max" is not advised.')
        self.add_option('-d', '--displays', dest='displays',
            help='comma-separated list of displays to operate on, or "auto" to detect; '+
                 'the first is the primary display.')
        self.add_option('-t', '--direction', dest='direction',
            choices=['left','right','top','bottom'],
            help='where to extend displays: "left", "right", "top", or "bottom"')
        self.add_option('', '--scaling', dest='scaling',
            choices=['default','native','scaled','centered','aspect-scaled'],
            help='flat-panel scaling mode: "default", "native", "scaled", "centered", or "aspect-scaled"')
        self.add_option('', '--plugins', dest='plugins',
            help='comma-separated list of plugins to enable. Special names: "user" for all user plugins '+
                 'in %s/hooks; "all" for all plugins found; "none" for no plugins.'%(
                 os.environ.get('XDG_CONFIG_HOME', os.path.join('~', '.config', 'disper'))))
        self.add_option('', '--cycle-stages', dest='cycle_stages', default='-c:-s:-S',
            help='colon-separated list command-line arguments to cycle through; "-S:-c:-s" by default')

        group = optparse.OptionGroup(self.parser, 'Actions',
            'Select exactly one of the following actions')
        self._add_option(group, '-l', '--list', action='append_const', const='list', dest='actions',
            help='list the attached displays')
        self._add_option(group, '-s', '--single', action='append_const', const='single', dest='actions',
            help='only enable the primary display')
        self._add_option(group, '-S', '--secondary', action='append_const', const='secondary', dest='actions',
            help='only enable the secondary display')
        self._add_option(group, '-c', '--clone', action='append_const', const='clone', dest='actions',
            help='clone displays')
        self._add_option(group, '-e', '--extend', action='append_const', const='extend', dest='actions',
            help='extend displays')
        self._add_option(group, '-p', '--export', action='append_const', const='export', dest='actions',
            help='export current settings to standard output')
        self._add_option(group, '-i', '--import', action='append_const', const='import', dest='actions',
            help='import current settings from standard input')
        self._add_option(group, '-C', '--cycle', action='append_const', const='cycle', dest='actions',
            help='cycle through the list of cycle stages')
        self.parser.add_option_group(group)


    def add_option(self, *args, **kwargs):
        '''adds an option to the parser. Implements append_const for Python<2.5 too'''
        return self._add_option(self.parser, *args, **kwargs)

    def _add_option(self, obj, *args, **kwargs):
        '''portable optarg add_option function that implements the append_const
        action for Python versions below 2.5; has an extra first argument as
        the object on which add_option should be called.'''
        if sys.hexversion < 0x020500f0 and 'action' in kwargs and \
                kwargs['action'] == 'append_const':
            # after: http://permalink.gmane.org/gmane.comp.python.optik.user/284
            def append_const_cb(const):
                def cb(opt, opt_str, value, parser):
                    if not getattr(parser.values, opt.dest):
                        setattr(parser.values, opt.dest, list())
                    getattr(parser.values, opt.dest).append(const)
                return cb
            kwargs['action'] = 'callback'
            kwargs['callback'] = append_const_cb(kwargs['const'])
            del kwargs['const']
        return obj.add_option(*args, **kwargs)


    def options_append(self, args):
        '''parses command-line options; can be called multiple times'''
        self.argv += args

    def options_parse(self, args=None):
        '''parses command-line options given; adds options to current list if set'''
        if args: self.options_append(args)
        (self.options, self.args) = self.parser.parse_args(self.argv)
        # need exactly one action
        if not self.options.actions: self.options.actions = []
        elif len(self.options.actions) > 1:
            self.parser.error('conflicting actions, please specify exactly one action: '
                         +', '.join(self.options.actions))
            raise SystemExit(2)

        if 'import' in self.options.actions or 'export' in self.options.actions:
            if self.options.resolution:
                self.log.warning('specified resolution ignored for %s'%self.options.actions[0])
            if self.options.displays:
                self.log.warning('specified displays ignored for %s'%self.options.actions[0])

        # apply defaults here to be able to detect if they were set explicitly or not
        if not self.options.direction: self.options.direction = "right"
        if not self.options.resolution: self.options.resolution = "auto"
        if not self.options.displays: self.options.displays = "auto"
        if not self.options.scaling: self.options.scaling = "default"
        if not self.options.debug: self.options.debug = logging.WARNING
        if self.options.plugins == None: self.options.plugins = "user"
        self.log.setLevel(self.options.debug)
        self.options.plugins = map(lambda x: x.strip(), self.options.plugins.split(','))
        if self.options.displays != 'auto':
            self.options.displays = map(lambda x: x.strip(), self.options.displays.split(','))
        if self.options.resolution not in ['auto', 'max', 'off']:
            self.options.resolution = map(lambda x: x.strip(), self.options.resolution.split(','))
        self.plugins.set_enabled(self.options.plugins)

    def config_read_default(self):
        '''Return default options from configuration files'''
        # Read old-style and XDG-style configuration files
        home = os.environ.get('HOME', '/')
        xdg_config_dirs = [os.path.join(home, '.disper')] + \
                          [os.environ.get('XDG_CONFIG_HOME', os.path.join(home, '.config', 'disper'))] + \
                          os.environ.get('XDG_CONFIG_DIRS', '/etc/xdg/disper').split(':')
        xdg_config_dirs = filter(lambda x: x and os.path.exists(x), xdg_config_dirs)
        # since later configuration files override previous ones, reverse order of reading
        # TODO allow override of action, since multiple actions would now conflict
        xdg_config_dirs = reversed(xdg_config_dirs)
        opts = ''
        for d in xdg_config_dirs:
            conffile = os.path.join(d, 'config')
            if not os.path.exists(conffile): continue
            f = open(conffile, 'r')
            opts = ''
            for l in f.readlines():
                opts += l.split('#',1)[0] + ' '
            f.close()
            # remember which configuration file was read last
            self.conffile = conffile
        return shlex.split(opts)

    def switch(self):
        '''Switch to configuration as specified in the options'''
        if len(self.options.actions) == 0:
            self.log.info('no action specified')
            # show help if no action specified
            self.parser.print_help()
            raise SystemExit(2)
        if 'single' in self.options.actions:
            self.switch_primary()
        elif 'secondary' in self.options.actions:
            self.switch_secondary()
        elif 'clone' in self.options.actions:
            self.switch_clone()
        elif 'extend' in self.options.actions:
            self.switch_extend()
        elif 'export' in self.options.actions:
            print self.export_config()
        elif 'import' in self.options.actions:
            self.import_config('\n'.join(sys.stdin))
        elif 'cycle' in self.options.actions:
            self._cycle(self.options.cycle_stages.split(':'))
        elif 'list' in self.options.actions:
            # list displays with resolutions
            displays = self.options.displays
            if displays == 'auto':
                displays = self.switcher().get_displays()
            for disp in displays:
                res = self.switcher().get_resolutions_display(disp)
                res.sort()
                print 'display %s: %s'%(disp, self.switcher().get_display_name(disp))
                print ' resolutions: '+str(res)
        else:
            self.log.critical('program error, unrecognised action: '+', '.join(self.options.actions))
            raise SystemExit(2)

    def switch_primary(self, res=None):
        '''Only enable primary display.
           @param res resolution to use; or 'auto' for default or None for option'''
        if self.options.displays and self.options.displays != 'auto':
            return self.switch_single(self.options.displays[0])
        else:
            return self.switch_single(self.switcher().get_primary_display())

    def switch_secondary(self, res=None):
        '''Only enable secondary display.
           @param res resolution to use; or 'auto' for default or None for option'''
        if self.options.displays and self.options.displays != 'auto':
            if len(self.options.displays)>=2:
                return self.switch_single(self.options.displays[1])
            else:
                self.log.critical('No secondary display found, falling back to primary.')
                return self.switch_single(self.options.displays[0])
        else:
            primary = self.switcher().get_primary_display()
            try:
                display = [x for x in self.switcher().get_displays() if x != primary][0]
            except IndexError:
                self.log.critical('No secondary display found, falling back to primary.')
                return self.switch_single(primary, res)
            return self.switch_single(display, res)

    def switch_single(self, display=None, res=None):
        '''Only enable one display.
           @param display display to enable; or 'auto' for primary or None for option
           @param res resolution to use; or 'auto' for default or None for option'''
        if not display: display = self.options.displays
        if display == 'auto':
            display = self.switcher().get_primary_display()
        elif isinstance(display, list) and len(display)>1:
            self.log.warning('single output requested but multiple specified; using first one')
            display = display[0]
        if display: display = [display]
        return self.switch_clone(display, res)

    def switch_clone(self, displays=None, res=None):
        '''Clone displays.
           @param displays list of displays; or 'auto' for default or None for option
           @param res resolution; or 'auto' for default, 'max' for max, 'off' to disable
                  the display, 'none' or None for option'''
        # figure out displays
        if not displays: displays = self.options.displays
        if displays == 'auto':
            displays = self.switcher().get_displays()
            self.log.info('auto-detected displays: '+', '.join(displays))
        else:
            self.log.info('using specified displays: '+', '.join(displays))
        # figure out resolutions
        if not res:
            res = self.options.resolution
            if type(res)==list or type(res)==tuple:
                if len(res) != 1: raise TypeError('need single resolution for clone')
                res = res[0]
        if res in ['auto', 'max']:
            r = self.switcher().get_resolutions(displays).common()
            if len(r)==0:
                self.log.critical('displays share no common resolution')
                raise SystemExit(1)
            if res == 'max': # ignore any preferred resolution
                for s in r: s.weight = 0
            res = sorted(r)[-1]
        else:
            res = Resolution(res)
        # and switch
        result = self.switcher().switch_clone(displays, res)
        self.plugins.set_layout_clone(displays, res)
        self.plugins.call('switch')
        return result

    def switch_extend(self, displays=None, direction=None, ress=None):
        '''Extend displays.
           @param displays list of displays; or 'auto for default or None for option
           @param direction direction to extend; or None for option
           @param ress list of resolutions; or 'auto' for default or 'max' for max or None for option'''
        # figure out displays
        if not displays: displays = self.options.displays
        if displays == 'auto':
            displays = self.switcher().get_displays()
            self.log.info('auto-detected displays: '+', '.join(displays))
        else:
            self.log.info('using specified displays: '+', '.join(displays))
        # figure out resolutions
        if not ress: ress = self.options.resolution
        if ress == 'max':     # max resolution for each
            # override auto-detection weights and get highest resolution
            ress = self.switcher().get_resolutions(displays)
            for rl in ress.values():
                for r in rl: r.weight = 0
            ress = ress.select()
            self.log.info('maximum resolutions for displays: '+str(ress))
        elif ress == 'auto':  # use preferred resolution for each
            ress = self.switcher().get_resolutions(displays).select()
            self.log.info('preferred resolutions for displays: '+str(ress))
        else:                       # list of resolutions specified
            ress = ResolutionSelection(ress, displays)
            if len(ress)==1:
                ress = ress * len(displays)
            elif len(ress) != len(displays):
                self.log.critical('resolution: must specify either "auto", "max", a single value, or one for each display')
                raise SystemExit(2)
            self.log.info('selected resolutions for displays: '+str(ress))
        # figure out direction
        if not direction: direction = self.options.direction
        # and switch
        result = self.switcher().switch_extend(displays, direction, ress)
        self.plugins.set_layout_extend(displays, direction, ress)
        self.plugins.call('switch')
        return result

    def export_config(self):
        return self.switcher().export_config()

    def import_config(self, data):
        result = self.switcher().import_config(data)
        self.plugins.call('switch')
        return result

    def _cycle(self, stages):
        # read last state
        stage = 0
        disperconf = None
        # TODO use X root window hint instead of file (which doesn't adhere to XDG)
        if self.conffile:
            disperconf = os.path.dirname(self.conffile)
        else:
            home = os.environ.get('HOME', '/')
            disperconf = os.environ.get('XDG_CONFIG_HOME', os.path.join(home, '.config', 'disper'))
        statefile = os.path.join(disperconf, 'last_cycle_stage')
        self.log.debug('Cycle state file: '+statefile)
        if os.path.exists(statefile):
            f = open(statefile, 'r')
            stage = int(f.readline())
            f.close()
        # apply next
        stage += 1
        if stage >= len(stages): stage = 0
        self.argv = filter(lambda x: x!='-C' and x!='--cycle', self.argv)
        self.options_parse(shlex.split(stages[stage]))
        try:
            self.switch()
        finally:
            # write new state to file; do it here to make sure that a
            # failing configuration doesn't block the cycling
            if not os.path.exists(disperconf):
                self.log.info('Creating disper configuration directory for statefile: '+disperconf)
                os.mkdir(disperconf)
            f = open(statefile, 'w')
            f.write(str(stage)+'\n')
            f.close()

    def switcher(self):
        '''Return switcher object (singleton).
        This is implemented as a method, so that it can be created only when
        needed to avoid errors when displaying help.'''
        if not self._switcher:
            self._switcher = Switcher()
        return self._switcher


def main():
    disper = Disper()
    disper.options_parse(sys.argv[1:])
    disper.switch()

if __name__ == "__main__":
    # Python 2.3 doesn't support arguments to basicConfig()
    try: logging.basicConfig(format='%(message)s')
    except: logging.basicConfig()
    main()

# vim:ts=4:sw=4:expandtab:



##############################################################################
# __init__.py - display switching with multiple backends
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#        
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License at http://www.gnu.org/licenses/gpl.txt
# By using, editing and/or distributing this software you agree to
# the terms and conditions of this license.

import logging

from edid import Edid
from resolutions import *

class Switcher:

    _displays = None
    _resolutions = ResolutionCollection()
    backend = None

    def __init__(self):
        '''Initialise the switcher and find a backend'''
        self.log = logging.getLogger('disper.switcher')
        self._probe_backend()

    def _probe_backend(self):
        '''Find and instantiate a suitable backend'''
        self.backend = None
        try:
            # nVidia must be probed before XRandR because it uses XRandR in a
            # non-standard way
            from swnvidia import NVidiaSwitcher
            self.backend = NVidiaSwitcher()
            self.log.info('backend: nVidia')
            return
        except SyntaxError: raise
        except: pass
        try:
            from swxrandr import XRandrSwitcher
            self.backend = XRandrSwitcher()
            self.log.info('backend: XRandR')
            return
        except SyntaxError: raise
        except: pass

        if not self.backend:
            raise Exception('No suitable backend found')

    ## the following methods must be defined by backends; see swnvidia.py
    ## for a complete example and an explanation of these methods
    #def get_displays(self):
    #def get_primary_display(self):
    #def get_display_name(self, ndisp):
    #def get_display_supported_res(self, ndisp):
    #def get_display_preferred_res(self, ndisp):
    #def get_display_edid(self, ndisp):
    #def switch_clone(self, displays, res):
    #def switch_extend(self, displays, direction, ress):
    #def import_config(self, cfg):
    #def export_config(self):

    def __getattr__(self, name):
        '''Pass unrecognised methods to the switcher itself; this is to
        simulate binding to a parent class at runtime.'''
        return getattr(self.backend, name)

    def get_displays(self):
        '''return an array of connected displays'''
        # hash displays to avoid probing twice
        if self._displays: return self._displays
        self._displays = self.backend.get_displays()
        # always put primary display in front
        if self.get_primary_display() in self._displays:
            self._displays = [self.get_primary_display()] + \
                filter(lambda x: x!=self.get_primary_display(), self._displays)
        return self._displays

    def get_resolutions_display(self, disp):
        '''return a list of resolutions for the specified display'''
        # hash resolutions to avoid probing them twice
        if disp in self._resolutions: return self._resolutions[disp]
        # get supported resolutions from driver
        r = ResolutionList(self.backend.get_display_supported_res(disp))
        if len(r)==0:
            r = ResolutionList('800x600, 640x480')
            self.log.warning('no resolutions found for display %s, falling back to: %s'%(disp, r))
        # bump weight of flat-panel display with 1000
        res = self.backend.get_display_preferred_res(disp)
        if res:
            res = Resolution(res)
            if res in r: r[r.index(res)].weight += 1000
            else: r.append(res)
        # bump weight of EDID resolutions with 100
        edid_data = self.backend.get_display_edid(disp)
        if edid_data:
            edid = Edid(edid_data)
            for d in edid.get_monitor_details():
                title, info = d
                if title != 'Detailed Timing': continue
                res = Resolution([info['horizontal_active'], info['vertical_active']])
                if res in r: r[r.index(res)].weight += 100
                else: r.append(res)
        self.log.info('resolutions of '+str(disp)+': '+', '.join(map(str,sorted(r))))
        self._resolutions[disp] = r
        return r

    def get_resolutions(self, displays):
        '''return a ResolutionCollection which is a hash with resolutions for
        each display'''
        res = ResolutionCollection()
        for disp in displays:
            res[disp] = self.get_resolutions_display(disp)
        return res

# vim:ts=4:sw=4:expandtab:



##############################################################################
# xnet.py - python versions of Xauth functions
#
# code taken from NvThermometer by Harry Organs, which was based on the code
# for python-xlib, written by Peter Liljenberg
# http://python-xlib.sourceforge.net/
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#        
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License at http://www.gnu.org/licenses/gpl.txt. 
# By using, editing and/or distributing this software you agree to 
# the terms and conditions of this license. 
#
# python procedural-style X protocol utilities
# based on GPL'ed code by Peter Liljenberg
# from python-xlib (http://python-xlib.sourceforge.net)
# i distilled this stuff from the classes in python-xlib
# to produce something a little more c-like.
# many thanks, saved me tons of digging through the
# xlib c code.


import socket
import os
import re
import struct
import string

import fcntl

if hasattr(fcntl, 'F_SETFD'):
    F_SETFD = fcntl.F_SETFD
    if hasattr(fcntl, 'FD_CLOEXEC'):
        FD_CLOEXEC = fcntl.FD_CLOEXEC
        FD_CLOEXEC = 1
else:
    from FCNTL import F_SETFD, FD_CLOEXEC

# protocol family constants from X.h provided by x.org
FamilyInternet  = 0
FamilyDECnet    = 1
FamilyChaos     = 2
FamilyInternet6 = 6

# constant for local hosts
FamilyLocal     = 256


###############################################################################
# a simple Exception class to raise X connection errors
#
class XConnectionError(Exception):

    def __init__(self,msg):
        self.msg= 'X CONNECTION ERROR: ' + msg

    def __str__(self):
        return self.msg


###############################################################################
# get X display info
#
__display_re = re.compile(r'^([-a-zA-Z0-9._]*):([0-9]+)(\.([0-9]+))?$')

def get_X_display(display=None):

    if display == None:
        d = os.environ.get('DISPLAY', '')
    else:
        d = display

    m = __display_re.match(d)
    if not m:
        raise XConnectionError( 'Bad display name' )
        
    host = m.group(1)
    if host == 'unix':
        host = ''

    dno = int(m.group(2))

    screen = m.group(4)
    if screen:
        screen = int(screen)
    else:
        screen = 0

    return d, host, dno, screen


###############################################################################
# return a socket connected to X server port
#
def get_X_socket(host,dno):
    s = None
    try:
        if host:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect((host, 6000 + dno))
        else:
            s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            s.connect('/tmp/.X11-unix/X%d' % dno)

    except socket.error, msg:
       s.close()
       raise XConnectionError( 'Socket error %s' % msg[1] )

    else:
        fcntl.fcntl(s.fileno(), F_SETFD, FD_CLOEXEC)

    return s

###############################################################################
# parse .Xauthority file
#
def parse_Xauthority( filename=None ):

    if filename is None:
        filename = os.environ.get('XAUTHORITY')

    if filename is None:
        try:
            filename = os.path.join(os.environ['HOME'], '.Xauthority')
        except KeyError:
            raise XConnectionError( "$HOME not set, can't find ~/.Xauthority" )

    try:
        xaf = open(filename, 'rb')
        raw = xaf.read()
        xaf.close()
    except IOError, err:
        raise XConnectionError( "Can't read ~/.Xauthority: %s" % err[1] )
        return []

    n = 0
    entries = []
    try:
        while n < len(raw):
            family, = struct.unpack('>H', raw[n:n+2])
            n = n + 2

            length, = struct.unpack('>H', raw[n:n+2])
            n = n + length + 2
            addr = raw[n - length : n]

            length, = struct.unpack('>H', raw[n:n+2])
            n = n + length + 2
            num = raw[n - length : n]

            length, = struct.unpack('>H', raw[n:n+2])
            n = n + length + 2
            name = raw[n - length : n]

            length, = struct.unpack('>H', raw[n:n+2])
            n = n + length + 2
            data = raw[n - length : n]

            if len(data) != length:
                break

            entries.append((family, addr, num, name, data))

    except struct.error, e:
        raise XConnectionError( '.Xauthority parsing failed' )

    if len(entries) == 0:
        raise XConnectionError( 'No connection authorization available' )

    return entries


###############################################################################
# find an authority to connect with
#
def match_X_auth(family, address, dispno, elist, types=("MIT-MAGIC-COOKIE-1",)):

    num = str(dispno)

    matches = {}

    for efam, eaddr, enum, ename, edata in elist:
        if efam == family and eaddr == address and num == enum:
            matches[ename] = edata

    for t in types:
        try:
            return (t, matches[t])
        except KeyError:
            return None


###############################################################################
# return the connection authority
#
def get_X_auth(sock, dname, host, dno):

    if host:
        family = FamilyInternet
        octets = string.split(sock.getpeername()[0], '.')
        addr = string.join(map(lambda x: chr(int(x)), octets), '')
    else:
        family = FamilyLocal
        addr = socket.gethostname()

    al = parse_Xauthority()
    auth = match_X_auth(family, addr, dno, al)
    
    if auth != None:
        return auth
    else:
        if family == FamilyInternet and addr == '\x7f\x00\x00\x01':
            family = FamilyLocal
            addr = socket.gethostname()
            return match_X_auth(family, addr, dno, al)


###############################################################################
# determine the byte order of architecture
#
def get_X_byteorder():

    bs=struct.unpack('BB', struct.pack('H', 0x0100))[0]
    if bs:
        order = 0x42
    else:
        order = 0x6c

    return order






##############################################################################
# swxrandr.py - display switching using XRandR
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#        
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License at http://www.gnu.org/licenses/gpl.txt
# By using, editing and/or distributing this software you agree to
# the terms and conditions of this license.

import logging

import xrandr

from resolutions import *

class XRandrSwitcher:

    def __init__(self):
        self.log = logging.getLogger('disper.switcher.xrandr')
        self.screen = xrandr.get_current_screen()
        if not xrandr.has_extension():
            raise Exception('No XRandR extension found')


    def get_displays(self):
        '''return an array of connected displays'''
        displays = self.screen.get_outputs()
        displays = filter(lambda o: o.is_connected(), displays)
        displays = map(lambda o: o.name, displays)
        return displays


    def get_primary_display(self):
        # no idea, just return first one for now
        return self.get_displays()[0]


    def get_display_name(self, ndisp):
        '''return the name of a display'''
        # nothing more as of now
        return ndisp


    def get_display_supported_res(self, ndisp):
        '''return a set of supported resolutions for a display.'''
        o = self.screen.get_output_by_name(ndisp)
        return o.get_available_resolutions()


    def get_display_preferred_res(self, ndisp):
        '''return the preferred resolution for a display.'''
        o = self.screen.get_output_by_name(ndisp)
        m = o.get_available_modes()[o.get_preferred_mode()]
        return [m.width,m.height]


    def get_display_edid(self, ndisp):
        '''return the EDID data for a display.'''
        # not available now
        return None


    def switch_clone(self, displays, res):
        '''switch to resolution and clone all displays'''
        ress = ResolutionSelection(res, displays)
        return self._switch(displays, ress, xrandr.RELATION_SAME_AS)


    def switch_extend(self, displays, direction, ress):
        '''extend desktop across all displays. direction is one of
        'left'/'right'/'bottom'/'top', and ress a dict of a resolution
        for each display.'''
        relation = None
        if direction=='left':
            relation = xrandr.RELATION_LEFT_OF
        elif direction=='right':
            relation = xrandr.RELATION_RIGHT_OF
        elif direction=='top':
            relation = xrandr.RELATION_ABOVE
        elif direction=='bottom':
            relation = xrandr.RELATION_BELOW
        else:
            raise ValueError('extend direction must be left/right/bottom/top')
        return self._switch(displays, ress, relation)


    def import_config(self, cfg):
        '''restore a display configuration as exported by export_config()'''
        raise NotImplementedError('import not yet implemented')


    def export_config(self):
        '''return a string that contains all information to set the current
        display configuration using import_config().'''
        raise NotImplementedError('export not yet implemented')


    def _switch(self, displays, ress, relation):
        '''switch displays to the specified resolution according to XRandR relation'''
        dprev = None
        old_displays = self.get_displays()
        for d in displays:
            res = ress[d]
            s = res.size()
            # for each display, select mode with highest refresh rate at res
            o = self.screen.get_output_by_name(d)
            modes = []
            for i,mode in enumerate(o.get_available_modes()):
                if mode.width != s[0]: continue
                if mode.height != s[1]: continue
                refresh = mode.dotClock/(mode.hTotal*mode.vTotal)
                modes.append([i, refresh])
            modes.sort(lambda x,y: x[1]-y[1])
            if len(modes) > 1:
                self.log.info(str(d)+': available refresh rates for resolution '+
                    str(res)+': '+', '.join(map(lambda o: '%d'%(o[1]), modes)))
            if len(modes) == 0:
                raise ValueError('Mode %dx%d is invalid for display %s'%(s[0], s[1], d))
            mode = modes[-1]
            self.log.info(str(d)+': selecting XRandR mode #%d: %s %dHz'%(mode[0],res,mode[1]))
            o.set_to_mode(mode[0])
            if dprev:
                o.set_relation(dprev, relation)
            dprev = d
            if d in old_displays: old_displays.remove(d)
            
        if len(old_displays)>0:
            for d in old_displays:
                o = self.screen.get_output_by_name(d)
                o.disable()
        
        self.screen.apply_output_config()

    def set_scaling(self, displays, scaling):
        if scaling == "default" : return
        raise NotImplementedError('scaling not implemented for XRandR')


# vim:ts=4:sw=4:expandtab:



###############################################################################
# nvcmd.py - nvidia NV-CONTROL X extension functions in python
#
# this file contains only a subset of the NV-CONTROL functions, expand and
# share when needed! See the nvidia-settings source for a complete C-based
# implementation.  This file contains high-level functions. If you need to
# add support for another command, please add it to this class.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#        
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License at http://www.gnu.org/licenses/gpl.txt
# By using, editing and/or distributing this software you agree to
# the terms and conditions of this license.

_all = ['GPU', 'Screen', 'NVidiaControl', 'metamode_clone', 'metamode_add_extend' ]

import re
from nvctrl import *
from nvctrl import NVidiaControl as NVidiaControlLowLevel
from metamodes import *

__BUS_TYPES = ['AGP', 'PCI', 'PCI Express', 'Integrated']
__OS_TYPES = ['Linux', 'FreeBSD', 'SunOS']
__ARCH_TYPES = ['x86', 'x86-64', 'IA64']


class NVidiaControl(NVidiaControlLowLevel):
    '''This class extends nvctrl.NVidiaControl with methods for
    accessing the NV-CONTROL functions on a higher level.'''

    def get_GPU_count(target):
        '''Return the number of GPU's present in the system.'''
        gpc = self.query_target_count(GPU())
        return gpc.count

    def get_bus_type(self, target):
        '''Return the bus type through which the GPU driving the specified X
        screen is connected to the computer.'''
        br = self.query_int_attribute(target, [], NV_CTRL_BUS_TYPE)
        return __BUS_TYPES[br.value]

    def get_OS_type(target):
        '''return the operating system on which the X server is running.'''
        ot = self.query_int_attribute(GPU(), [], NV_CTRL_OPERATING_SYSTEM )
        return __OS_TYPES[ot.value]

    def get_host_architecture(target):
        '''return the architecture on which the X server is running.'''
        ha = self.query_int_attribute(GPU(), NV_CTRL_TARGET_TYPE_X_SCREEN, [], NV_CTRL_ARCHITECTURE )
        return __ARCH_TYPES[ha.value]

    def get_vram(self, target):
        '''Return the total amount of memory available to the specified GPU
        (or the GPU driving the specified X screen). Note: if the GPU supports
        TurboCache(TM), the value reported may exceed the amount of video
        memory installed on the GPU. The value reported for integrated GPUs may
        likewise exceed the amount of dedicated system memory set aside by the
        system BIOS for use by the integrated GPU.'''
        vr = self.query_int_attribute(target, NV_CTRL_TARGET_TYPE_GPU, [], NV_CTRL_VIDEO_RAM)
        return vr.value

    def get_IRQ(self, target):
        '''Return the interrupt request line used by the GPU driving the screen'''
        irq = self.query_int_attribute(target, NV_CTRL_TARGET_TYPE_GPU, [], NV_CTRL_IRQ)
        return irq.value

    def get_connected_displays(self, target):
        '''Return an array with connected display numbers'''
        cd = self.query_int_attribute(target, [], NV_CTRL_CONNECTED_DISPLAYS)
        return self._mask2displays(cd.value)

    def get_enabled_displays(self, target):
        '''returns an array of displays that are enabled on the specified X
        screen or GPU.'''
        ed = self.query_int_attribute(target, [], NV_CTRL_ENABLED_DISPLAYS)
        return self._mask2displays(ed.value)

    def supports_framelock(self, target):
        '''returns whether the underlying GPU supports Frame Lock. All of the
        other frame lock attributes are only applicable if this returns True.'''
        fl = self.query_int_attribute(target, [], NV_CTRL_FRAMELOCK)
        return fl.value == 1

    def get_name(self, target):
        '''the GPU product name on which the specified X screen is running'''
        ns = self.query_string_attribute(target, [], NV_CTRL_STRING_PRODUCT_NAME)
        return ns.string

    def get_driver_version(self, target):
        '''the NVIDIA (kernel level) driver version for the specified screen or GPU'''
        ns = self.query_string_attribute(target, [], NV_CTRL_STRING_NVIDIA_DRIVER_VERSION)
        return ns.string

    def get_vbios_version(self, target):
        '''the version of the VBIOS for the specified screen or GPU'''
        ns = self.query_string_attribute(target, [], NV_CTRL_STRING_VBIOS_VERSION)
        return ns.string

    def get_current_modeline(self, target):
        '''Return the ModeLine currently being used by the specified display
        device. This attribute may be queried using a GPU() or Screen() target.
        The ModeLine string may be prepended with a comma-separated list of
        "token=value" pairs, separated from the ModeLine string by "::". This
        "token=value" syntax is the same as that used in TODO'''
        dm = self.query_string_attribute( target, [], NV_CTRL_STRING_CURRENT_MODELINE)
        return dm.string

    def GVO_supported(target):
        '''returns whether this X screen supports GVO; if this screen does not
        support GVO output, then all other GVO attributes are unavailable.'''
        gv = self.query_int_attribute(Screen(), [], NV_CTRL_GVO_SUPPORTED)
        return gv.value == 1

    def get_core_temp(self, target):
        '''return the current core temperature of the GPU driving the X screen.'''
        ct = self.query_int_attribute(target, [], NV_CTRL_GPU_CORE_TEMPERATURE)
        return ct.value

    def get_core_threshold(self, target):
        '''return the current GPU core slowdown threshold temperature. It
        reflects the temperature at which the GPU is throttled to prevent
        overheating.'''
        th = self.query_int_attribute(target, [], NV_CTRL_GPU_CORE_THRESHOLD)
        return th.value

    def get_default_core_threshold(self, target):
        '''return the default core threshold temperature.'''
        dt = self.query_int_attribute(target, [], NV_CTRL_GPU_DEFAULT_CORE_THRESHOLD)
        return dt.value

    def get_max_core_threshold(self, target):
        '''return the maximum core threshold temperature.'''
        mt = self.query_int_attribute(target, [], NV_CTRL_GPU_MAX_CORE_THRESHOLD)
        return mt.value

    def get_ambient_temp(self, target):
        '''return the current temperature in the immediate neighbourhood of
        the GPU driving the X screen.'''
        at = self.query_int_attribute(target, [], NV_CTRL_AMBIENT_TEMPERATURE)
        return at.value

    def get_2D_clocks(self, target):
        '''query the 2D (GPU, memory) clocks of the device driving the X screen.
        All clock values are in MHz.'''
        cl = self.query_int_attribute(target, [], NV_CTRL_GPU_2D_CLOCK_FREQS)
        return (cl.value >> 16, cl.value & 0xFFFF)

    def get_3D_clocks(self, target):
        '''query the 3D (GPU, memory) clocks of the device driving the X screen.
        All clock values are in MHz.'''
        cl = self.query_int_attribute(target, [], NV_CTRL_GPU_3D_CLOCK_FREQS)
        return (cl.value >> 16, cl.value & 0xFFFF)

    def get_default_2D_clocks(self, target):
        '''query the default 2D (GPU, memory) clocks of the device driving the
        X screen. All clock values are in MHz.'''
        cl = self.query_int_attribute(target, [], NV_CTRL_GPU_DEFAULT_2D_CLOCK_FREQS)
        return (cl.value >> 16, cl.value & 0xFFFF)

    def get_default_3D_clocks(self, target):
        '''query the default 3D (GPU, memory) clocks of the device driving the
        X screen. All clock values are in MHz.'''
        cl = self.query_int_attribute(target, [], NV_CTRL_GPU_DEFAULT_3D_CLOCK_FREQS)
        return (cl.value >> 16, cl.value & 0xFFFF)

    def get_current_clocks(self, target):
        '''return the current (GPU, memory) clocks of the graphics device
        driving the X screen.'''
        cl = self.query_int_attribute(target, [], NV_CTRL_GPU_DEFAULT_CLOCK_FREQS)
        return (cl.value >> 16, cl.value & 0xFFFF)

    def get_xinerama_enabled(self, target):
        '''return whether Xinerama is enabled or not'''
        xn = self.query_int_attribute(target, [], NV_CTRL_XINERAMA)
        return xn.value == 1

    def get_refresh_rate(self, target, display):
        '''return the refresh rate of the specified display device in Hz.'''
        rr = self.query_int_attribute(target, [display], NV_CTRL_REFRESH_RATE)
        return rr.value / 100.0

    def probe_displays(self, target):
        '''re-probes the hardware to detect what display devices are connected
        to the GPU or GPU driving the specified X screen. Returns an array
        of displays.'''
        md = self.query_int_attribute(target, [], NV_CTRL_PROBE_DISPLAYS)
        return self._mask2displays(md.value)

    def get_display_name(self, target, display):
        '''return the name of the display device number.'''
        ns = self.query_string_attribute(target, [display], NV_CTRL_STRING_DISPLAY_DEVICE_NAME)
        return ns.string

    def get_display_modelines(self, target, display):
        '''return a display device's supported ModeLines as an array of
        ModeLine strings. The attribute may be queried through a GPU() or
        Display().

        Each ModeLine string may be prepended with a comma-separated list
        of token=value pairs, separated from the ModeLine string with a
        "::".'''
        mls = self.query_binary_data(target, [display], NV_CTRL_BINARY_DATA_MODELINES)
        return filter(lambda x: x, mls.data.split('\0'))

    def get_current_metamode(self, target):
        '''returns the metamode currently being used by the specified X
        screen. The MetaMode string has the same syntax as the MetaMode
        X configuration option, as documented in the NVIDIA driver README.

        The returned string may be prepended with a comma-separated list of
        "token=value" pairs, separated from the MetaMode string by "::".
        This "token=value" syntax is the same as that used in
        get_metamodes().'''
        mm = self.query_string_attribute(target, [], NV_CTRL_STRING_CURRENT_METAMODE)
        return MetaMode(mm.string)

    def get_metamodes(self, target):
        '''return an X screen's supported MetaModes as an array of MetaMode
        strings. The attribute must be queried through a Display().
        Each MetaMode string may be prepended with a comma-separated list
        of token=value paris, separated from the  MetaMode string with
        "::".'''
        mms = self.query_binary_data(target, [], NV_CTRL_BINARY_DATA_METAMODES)
        return MetaModeList(filter(lambda x: x, mms.data.split('\0')))

    def add_metamode(self, target, mm):
        '''provide a MetaMode string as input. return newly created MetaMode
        on success, or -1 if it failed.

        All ModeLines referenced in the MetaMode must already exist for each
        display device (as returned by get_metamodes()).
        
        The MetaMode string should have the same syntax as the MetaMode X
        configuration option, as documented in the NVIDIA driver README.
        
        The input string can optionally be prepended with a string of
        comma-separated "token=value" pairs, separated from the MetaMode string
        by "::". Currently, the only valid token is "index" which indicates the
        insertion index for the MetaMode.'''
        #res = self.set_string_attribute(target, [], NV_CTRL_STRING_ADD_METAMODE, mm)
        res = self.string_operation(target, [], NV_CTRL_STRING_OPERATION_ADD_METAMODE, str(mm))
        r = re.match('id=(\d+)', res.string)
        if not r: return -1
        return int(r.group(1))

    def delete_metamode(self, target, mm):
        '''Deletes an existing MetaMode from the specified X Screen. The
        currently selected MetaMode cannot be deleted. (This also means you
        cannot delete the last MetaMode).  The MetaMode string should have the
        same syntax as the MetaMode X configuration option, as documented in
        the NVIDIA driver README.

        The argument can either be a metamode, MetaMode, or a mode id (integer).'''
        if type(mm) == int:
            # retrieve id from MetaModes
            mm = self.get_metamodes(target).find(mm)
            if not mm:
                # not found, modeline id not found
                return False
            mm = mm.src
        elif isinstance(mm, MetaMode):
            # get original string from metamode
            mm = mm.src
        mm = re.sub(r'^.*::\s*', r'', mm)
        res = self.set_string_attribute(target, [], NV_CTRL_STRING_DELETE_METAMODE, mm)
        return res.flags

    def move_metamode(self, target, mm, to):
        '''moves a MetaMode to the specified index location.
        The MetaMode must already exist in the X Screen's list of MetaModes (as
        returned by the get_metamodes(). If the index is larger than the number
        of MetaModes in the list, the MetaMode is moved to the end of the list.
        The MetaMode string should have the same syntax as the MetaMode X
        configuration option, as documented in the NVIDIA driver README.
 
        The MetaMode string must be prepended with a comma-separated list
        of "token=value" pairs, separated from the MetaMode string by "::".
        Currently, the only valid token is "index", which indicates where
        in the MetaMode list the MetaMode should be moved to.
        
        Other tokens may be added in the future.
        
        E.g.,
          "index=5 :: CRT-0: 1024x768 @1024x768 +0+0"
        
        The mm argument can either be a metamode, MetaMode, or a mode id (integer).
        '''
        if type(mm) == int:
            # retrieve id from MetaModes
            mm = self.get_metamodes(target).find(mm)
            if not mm:
                # not found, modeline id not found
                return False
            mm = mm.src
        elif isinstance(mm, MetaMode):
            # get original string from metamode
            mm = mm.src
        mm = "index=%d :: %s"%(to, re.sub(r'^.*::\s*', r'', mm))
        res = self.set_string_attribute(target, [], NV_CTRL_STRING_MOVE_METAMODE, mm)
        return res.flags
         

    def build_display_modepool(self, target, display, opt=None):
        '''build a ModePool for the specified display device on the specified
        target (either an X screen or a GPU). This is typically used to
        generate a ModePool for a display device on a GPU on which no X screens
        are present.
        
        Currently, a display device's ModePool is static for the life of the X
        server, so this will return False if requested to build a ModePool on a
        display device that already has a ModePool.
        
        The string input opt may be NULL. If it is not NULL, then it is
        interpreted as a double-semicolon ("::") separated list of
        "option=value" pairs, where the options and the syntax of their values
        are the X configuration options that impact the behavior of modePool
        construction; namely: "ModeValidation" "HorizSync" "VertRefresh"
        "FlatPanelProperties" "TVStandard" "ExactModeTimingsDVI" "UseEdidFreqs"
        
        An example input string might look like:
        "ModeValidation=NoVesaModes :: HorizSync=50-110 :: VertRefresh=50-150"
        
        Returns a boolean.'''
        res = self.string_operation(target, [display], NV_CTRL_STRING_OPERATION_BUILD_MODEPOOL, opt)
        return res.flags

    def get_screen_associated_displays(self, target):
        '''return array of display devices that are "associated" with the
        specified X screen.'''
        res = self.query_int_attribute(target, [], NV_CTRL_ASSOCIATED_DISPLAY_DEVICES)
        return self._mask2displays(res.value)

    def set_screen_associated_displays(self, target, displays):
        '''set which display devices are "associated" with the specified X
        screen (ie: are available to the X screen for displaying the X
        screen).'''
        mask = self._displays2mask(displays)
        res = self.set_int_attribute(target, [], NV_CTRL_ASSOCIATED_DISPLAY_DEVICES, mask)
        return res.flags

    def get_frontend_resolution(self, target, display):
        '''return the dimensions of the frontend (current) resolution as
        determined by the NVIDIA X Driver as [width, height].'''
        res = self.query_int_attribute(target, [display], NV_CTRL_FRONTEND_RESOLUTION)
        if not res.flags: return False
        return res>>16, res&0xffff

    def get_backend_resolution(self, target, display):
        '''return the dimensions of the backend resolution as determined by the
        NVIDIA X Driver as [width, height].

        The backend resolution is the resolution (supported by the display
        device) the GPU is set to scale to.  If this resolution matches the
        frontend resolution, GPU scaling will not be needed/used.'''
        res = self.query_int_attribute(target, [display], NV_CTRL_BACKEND_RESOLUTION)
        if not res.flags: return False
        return res.value>>16, res.value&0xffff

    def get_dfp_native_resolution(self, target, display):
        '''return the dimensions of the native resolution of the flat panel as
        determined by the NVIDIA X Driver as [width, height].

        The native resolution is the resolution at which a flat panel
        must display any image.  All other resolutions must be scaled to this
        resolution through GPU scaling or the DFP's native scaling capabilities
        in order to be displayed.

        This attribute is only valid for flat panel (DFP) display devices.'''
        res = self.query_int_attribute(target, [display], NV_CTRL_FLATPANEL_NATIVE_RESOLUTION)
        if not res.flags: return False
        return res.value>>16, res.value&0xffff

    def get_dfp_best_fit_resolution(self, target, display):
        '''return the dimensions of the resolution, selected by the X driver,
        from the DFP's EDID that most closely matches the frontend resolution
        of the current mode as [width, height]. The best fit resolution is
        selected on a per-mode basis. set_scaling() is used to select between
        get_best_fit_resolution() and get_native_resolution().

        This attribute is only valid for flat panel (DFP) display devices.'''
        res = self.query_int_attribute(target, [display], NV_CTRL_FLATPANEL_BEST_FIT_RESOLUTION)
        if not res.flags: return False
        return res.value>>16, res.value&0xffff

    def set_gpu_scaling(self, target, display, starget, smethod):
        '''controls what the GPU scales to and how. If starget is 'best-fit',
        the GPU scales the frontend (current) mode to the closest larger
        resolution in the flat panel's EDID and allow the flat panel to do its
        own scaling to the native resolution. If starget is 'native', the GPU
        scales the frontend (current) mode to the flat panel's native
        resolution, thus disabling any internal scaling the flat panel might
        have.
        
        smethod can be 'strechted', 'centered' or 'aspect-scaled'.
        
        starget and smethod can also be numbers.'''
        mode = 0
        if type(starget) == int:
            mode += (starget&0xffff)<<16
        elif starget == 'native':
            mode += 1<<16
        elif starget == 'best-fit':
            mode += 2<<16
        else:
            raise ValueError('Scaling target must be either "best fit" or "native"')
        if type(smethod) == int:
            mode += smethod & 0xffff
        elif smethod == 'stretched':
            mode += 1
        elif smethod == 'centered':
            mode += 2
        elif smethod == 'aspect-scaled':
            mode += 3
        else:
            raise ValueError('Scaling method must be one of "stretched", "centered" or "aspect-scaled"')
        res = self.set_int_attribute(target, [display], NV_CTRL_GPU_SCALING, mode)
        return res.flags

    def get_gpu_scaling(self, target, display):
        '''return current GPU scaling as [target, method].
        See set_gpu_scaling() for details.'''
        res = self.query_int_attribute(target, [display], NV_CTRL_GPU_SCALING)
        if not res.flags: return False
        starget = res.value >> 16
        if starget == 1:   starget = 'native'
        elif starget == 2: starget = 'best-fit'
        smethod = res.value & 0xffff
        if smethod == 1:   smethod = 'stretched'
        elif smethod == 2: smethod = 'centered'
        elif smethod == 3: smethod = 'aspect-scaled'
        return starget, smethod

    def get_max_displays(self, target):
        '''return the maximum number of display devices that can be driven
        simultaneously on a GPU (e.g., that can be used in a MetaMode at once).
        Note that this does not indicate the maximum number of bits that can be
        set in NV_CTRL_CONNECTED_DISPLAYS, because more display devices can be
        connected than are actively in use.'''
        res = self.query_int_attribute(target, [], NV_CTRL_MAX_DISPLAYS)
        if not res.flags: return False
        return res.value

    def get_display_edid(self, target, display):
        '''return a display device's EDID information data'''
        mms = self.query_binary_data(target, [display], NV_CTRL_BINARY_DATA_EDID)
        if not mms.flags: return False
        return mms.data
    
    def get_xinerama_info_order(self, target):
        '''return the order that display devices will be returned via
        Xinerama when TwinViewXineramaInfo is enabled.  Follows the same
        syntax as the TwinViewXineramaInfoOrder X config option.'''
        res = self.query_string_attribute(target, [], NV_CTRL_STRING_TWINVIEW_XINERAMA_INFO_ORDER)
        if not res.flags: return False
        return res.string
        
    def set_xinerama_info_order(self, target, displays):
        '''specify the order that display devices will be returned via
        Xinerama when TwinViewXineramaInfo is enabled.  Follows the same
        syntax as the TwinViewXineramaInfoOrder X config option.'''
        res = self.set_string_attribute(target, [], NV_CTRL_STRING_TWINVIEW_XINERAMA_INFO_ORDER, ', '.join(displays))
        return res.flags


# vim:ts=4:sw=4:expandtab:



##############################################################################
# switcher-nvidia.py - display switching for nVidia cards
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#        
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License at http://www.gnu.org/licenses/gpl.txt
# By using, editing and/or distributing this software you agree to
# the terms and conditions of this license.

import os
import re
import logging

import nvidia

from resolutions import *

class NVidiaSwitcher:

    nv = None
    _display_associations = []
    _switch_method = None


    def __init__(self):
        self.nv = nvidia.NVidiaControl()
        self.screen = nvidia.Screen(self.nv.xscreen)
        self.log = logging.getLogger('disper.switcher.nvidia')
        # new NVidia driver versions can just use XRandR, so fail in that
        # case to give the other driver a chance
        dversion = self.nv.get_driver_version(self.screen)
        self.log.info('NVidia kernel driver version: %s'%dversion)
        if int(dversion.split('.')[0]) >= 300:
            raise Exception('NVidia driver >= 300 uses XRandR 1.2')
        # either use XRandR module or command-line utility
        try:
            import xrandr
            self._switch_method=self._xrandr_switch_mod
        except:
            self.log.info('using xrandr command instead of XRandR module')
            self._switch_method=self._xrandr_switch_cmd


    def get_displays(self):
        '''return an array of connected displays'''
        displays = self.nv.probe_displays(self.screen)
        return displays


    def get_primary_display(self):
        '''return the primary display of this system. I'm not really sure how
        to do this, so currently the first found flat panel display (DFP) is
        returned, or CRT if none, or TV if neither.'''
        displays = self.get_displays()
        for d in ['DFP', 'CRT', 'TV']:
            for i in range(8):
                disp = '%s-%1d'%(d,i)
                if disp in displays:
                    return disp
        # this should be unreachable code, but be safe
        self.log.error('program error, could not determine primary display')
        return displays[0]


    def get_display_name(self, ndisp):
        '''return the name of a display'''
        return self.nv.get_display_name(self.screen, ndisp)


    def get_display_supported_res(self, ndisp):
        '''return a set of supported resolutions for a display.
        Displays need to be associated to probe their modelines, so this method
        temporarily changes that (and reverts to the old setup before
        returning).'''

        # Get display resolutions for display. The display needs to have
        # it associated to the X screen to be able to do this. So we check that
        # it is associated first and do so when needed. Restore associated
        # displays afterwards.
        # Note: When twinview has not been enabled before, the X server can
        #       *crash* when a display is associated that isn't mentioned in
        #       any metamode line. So create an autoselect modeline first.
        self._push_display_association([ndisp])
        try:
            self.nv.build_display_modepool(self.screen, ndisp)
            resolutions = set()
            for m in self.nv.get_display_modelines(self.screen, ndisp):
                r = re.search(r'::\s*"(\d+x\d+)"', m)
                if not r: continue
                resolutions.add(r.group(1))
        finally:
            self._pop_display_association()

        return resolutions


    def get_display_preferred_res(self, ndisp):
        '''return the preferred resolution for a display.
        Displays need to be associated to probe their modelines, so this method
        temporarily changes that (and reverts to the old setup before
        returning).'''
        self._push_display_association([ndisp])
        try:
            self.nv.build_display_modepool(self.screen, ndisp)
            res = self.nv.get_dfp_native_resolution(self.screen, ndisp)
        finally:
            self._pop_display_association()
        return res


    def get_display_edid(self, ndisp):
        '''return the EDID data for a display.'''
        return self.nv.get_display_edid(self.screen, ndisp)


    def switch_clone(self, displays, res):
        '''switch to resolution and clone all displays'''
        mm = nvidia.metamode_clone(displays, str(res))
        return self._switch(mm, displays)


    def switch_extend(self, displays, direction, ress):
        '''extend desktop across all displays. direction is one of
        'left'/'right'/'bottom'/'top', and ress a dict of a resolution
        for each display.'''
        mm = None
        for disp in displays:
            res = ress[disp]
            mm = nvidia.metamode_add_extend(mm, direction, disp, str(res))
        return self._switch(mm, displays)


    def import_config(self, cfg):
        '''restore a display configuration as exported by export_config()'''
        backend = displays = mmline = scaling = xio = None
        for l in cfg.splitlines():
            key, sep, value = map(lambda s: s.strip(), l.partition(':'))
            if key == 'backend':
                backend = value
            elif key == 'associated displays':
                displays = map(lambda s: s.strip(), value.split(','))
            elif key == 'metamode':
                mmline = value.strip()
            elif key == 'scaling':
                scaling = map(lambda s: s.strip(), value.split(','))
            elif key == 'xinerama info order':
                xio = map(lambda s: s.strip(), value.split(','))

        if not backend:
            self.log.warning('no backend specified, assuming nvidia')
        elif backend != 'nvidia':
            raise Exception('can only import what is exported by the nvidia backend')
        if not displays:
            raise Exception('"associated displays" missing from configuration')
        if not mmline:
            raise Exception('"metamode" missing from configuration')
        # scaling, xio is optional

        self._switch(mmline, displays, xio=xio)

        if scaling:
            if len(scaling) != len(displays):
                raise Exception('number of entries in "scaling" must equal # associated displays')
            self.set_scaling(displays, scaling)
            

    def export_config(self):
        '''return a string that contains all information to set the current
        display configuration using import_config().'''
        cfg = ['backend: nvidia']
        # associated displays
        assocdisplays = self.nv.get_screen_associated_displays(self.screen)
        cfg.append('associated displays: ' + ', '.join(assocdisplays))
        # and current metamode; remove options since they can't be specified
        # when restoring
        mm = self.nv.get_current_metamode(self.screen)
        mm = re.sub(r'^.*::\s*', r'', str(mm))
        cfg.append('metamode: ' + mm)
        # scaling mode of displays
        scalings = self.get_scaling(assocdisplays)
        cfg.append('scaling: ' + ', '.join(scalings))
        # Xinerama info order
        xio = filter(lambda x: x in assocdisplays, 
            map(lambda s: s.strip(), 
            self.nv.get_xinerama_info_order(self.screen).split(',')))
        cfg.append('xinerama info order: ' + ", ".join(xio))
        return '\n'.join(cfg)


    def _switch(self, mmline, displays, scaling=None, xio=None):
        '''switch to the specified metamode. mmline is a MetaMode, displays
        a list of displays to use, and scaling an optional argument specifying
        the gpu scaling, which is either an array of two elements as returned
        by NVidiaControl.get_gpu_scaling(), or an array of these elements, one
        for each display.'''
        
        '''the nvidia-settings source mentions the following in function
        update_screen_metamodes() defined in ctkdisplayconfig.c
        
        * preprocess
          - get current list of metamodes for this screen
          - add all new metamodes to the end of the list
        * mode switch if needed
        * postprocess
          - delete any unused mode
          - move metamodes to correct location
          
        pseudocode:
          metamode_strs = nv.get_current_list_of_metamodes()
          cur_metamode_str = nv.get_current_metamode()
          preprocess_metamodes(screen, metamode_strs) = {
            for metamode in screen.metamodes:
              if metamode in metamode_strs:
                metamode_strs.find(metamode).keep = True
              else:
                nv.add_metamode(metamode)
          }
          if metamode_str == screen->cur_metamode: clear_apply = 1
          postprocess_metamodes(screen, metamode_strs) = {
            for metamode in metamode_strs:
              if not metamode.keep:
                nv.delete_metamode(metamode)
            order_metamodes(screen) = {
              for i in len(metamodes):
                nv.move_metamode(metamode[i], i) 
            }
          }
        '''

        # make sure requested displays are connected (or metamode can't be created)
        unconndisplays = filter(lambda x: x not in self.get_displays(), displays)
        if len(unconndisplays) > 0:
            raise Exception('unconnected displays referenced, please connect: ' + \
                ', '.join(unconndisplays))

        # default xinerama info order
        if not xio:
            xio = displays

        # find or create MetaMode
        oldxio = self.nv.get_xinerama_info_order(self.screen)
        self._push_display_association(displays)
        try:
            mm = self.nv.get_metamodes(self.screen).find(mmline)
            if mm:
                mmid = mm.id
            else:
                mmid = self._add_metamode(mmline)
                if mmid < 0:
                    raise Exception('could not find nor create MetaMode: %s'%mmline)
                
            # set order before switch so it is detected
            self.log.info('setting xinerama info order: '+ ', '.join(xio))
            self.nv.set_xinerama_info_order(self.screen, xio)

            # change to this mode using xrandr and refresh as id
            self._xrandr_switch(mmid)
        except:
            # delete dangling metamodes and deassociate old
            self._cleanup_metamodes(displays)
            self._pop_display_association()
            self.nv.set_xinerama_info_order(self.screen, oldxio)
            raise

        # delete dangling metamodes and deassociate old
        self._cleanup_metamodes(displays)
        self._pop_display_association(False)
        self._set_associated_displays(displays)


    def _add_metamode(self,  mm):
        '''add a metamode. Returns id of newly created metamode, or -1 if it
        already existed.'''
        ## To enter a MetaMode line, the displays involved must have been
        ## associated or the nvidia driver doesn't remember display names.
        self.log.info('adding metamode: %s'%mm)
        return self.nv.add_metamode(self.screen, mm)


    def _add_metamode_autoselect(self, displays):
        '''add a temporary auto-select metamode that is needed before
        associating displays. returns id of metamode created, or -1 if it
        already existed'''
        ## There must be a metamode containing all displays when associating
        ## displays, or the X server may crash.
        mm = nvidia.metamode_clone(displays, 'nvidia-auto-select')
        self.log.info('adding auto-select metamode: %s'%str(mm))
        return self.nv.add_metamode(self.screen, mm)


    def _delete_metamode(self, id):
        '''delete the specified metamode.'''
        self.log.info('deleting metamode: %d'%id)
        return self.nv.delete_metamode(self.screen, id)


    def _set_associated_displays(self, displays):
        '''set the displays associated to the current X screen. Don't use this
        function directly, rather use both
        _push_display_association() and _pop_display_association().
        The primary display is kept associated always to avoid problems when
        switching from one single display device to another. [bug #315920]'''
        self.log.info('associating displays: %s'%(', '.join(displays)))
        d = set(displays)
        d.add(self.get_primary_display())
        return self.nv.set_screen_associated_displays(self.screen, d)


    def _push_display_association(self, displays):
        '''add a display to the currently associated displays and save
        previous state to return to using _pop_display_association().
        It is best to put the latter in a try ... finally clause to make
        sure it is always called.'''
        # change association when needed
        olddisplays = self.nv.get_screen_associated_displays(self.screen)
        assocdisplays = set(olddisplays).union(set(displays))
        if set(olddisplays) != set(assocdisplays):
            oldid = self._add_metamode_autoselect(assocdisplays)
            self._set_associated_displays(assocdisplays)
            self._display_associations.append((olddisplays, oldid))
        else:
            self._display_associations.append((None, -1))


    def _pop_display_association(self, dorestore = True):
        '''restore the previous display association from an earlier
        _push_display_association() call. If dorestore is False, only the old
        values will be popped and the display will not be restored, which is
        useful when the association is overridden afterwards.'''
        if len(self._display_associations) == 0:
            raise Exception('unbalanced _pop_display_assocation(), this is a program bug')
        olddisplays, oldid = self._display_associations.pop()
        # restore when needed
        if not olddisplays: return
        if oldid > 0:
            self._delete_metamode(oldid)
        if dorestore:
            self._set_associated_displays(olddisplays)


    def _xrandr_switch(self, mmid, virtualres=None):
        '''switch to the specified MetaMode id. Also the virtual resolution is
        needed; this will be retrieved when omitted.

        The virtual resolution is needed when a MetaMode is added and removed.
        I suspect that the XRandR refresh rates are not removed propely by the
        nVidia driver.'''
        if not virtualres:
            mm = self.nv.get_metamodes(self.screen).find(mmid)
            virtualres = mm.bounding_size()
        return self._switch_method(mmid, virtualres)

    def _xrandr_switch_mod(self, mmid, virtualres):
        '''_xrandr_switch that uses the xrandr python module'''
        import xrandr
        screen = xrandr.get_current_screen()
        sizeidx = -1
        for i,s in enumerate(screen.get_available_sizes()):
            if s.width != virtualres[0] or s.height != virtualres[1]: continue
            if mmid in screen.get_available_rates_for_size_index(i):
                res = '%dx%d' % ( s.width, s.height )
                sizeidx = i
                break
        if sizeidx < 0:
            raise Exception( 'could not switch to metamode %d: resolution not found' % mmid )
        screen.set_size_index(sizeidx)
        screen.set_refresh_rate(mmid)
        logging.info('switching to metamode %d: [%d] %s / %s'%(mmid,sizeidx,res,mmid))
        screen.apply_config()

    def _xrandr_switch_cmd(self, mmid, virtualres):
        '''_xrandr_switch that uses the command 'xrandr' '''
        cmd='xrandr -s %dx%d -r %d'%(virtualres[0],virtualres[1],mmid)
        logging.info('switching to metamode %d: %s'%(mmid,cmd))
        return os.system(cmd)

    def _cleanup_metamodes(self, displays):
        '''cleanup metamodes referencing displays that are not associated.
        driver loses display names in metamodes when displays are not
        associated and they remain around. Some more details to be found on
          http://www.nvnews.net/vbulletin/showthread.php?t=123781
        It is important that all displays currently present in the metamode
        list are still associated to the X screen.
        '''
        metamodes = self.nv.get_metamodes(self.screen)
        for mm in metamodes:
            for d in mm.metamodes:
                if d.display not in displays and d.physical:
                    self.log.info('deleting dangling metamode %d: %s'%(mm.id,mm))
                    r = self.nv.delete_metamode(self.screen, mm)
                    if not r: self.log.warning('deletion of dangling metamode %d failed'%mm.id)
                    break
        # nvidia-settings re-orders them, so do it here to be sure
        metamodes = self.nv.get_metamodes(self.screen)
        for i,mm in enumerate(metamodes):
            self.nv.move_metamode(self.screen, mm, i)

    def set_scaling(self, displays, scaling):
        '''update the flat panel scaling mode if it was set previously by
        nvidia-settings. scaling must be one of: default, native, scaled, centered,
        aspect-scaled. Alternatively, scaling can be a list specifying the scaling
        for each display separately.
        The default choice parses the nvidia-settings configuration file
        ~/.nvidia-settings-rc to obtain the default value, or does nothing if that
        fails. Note that nvidia-settings may or may not save this information from
        the gui; both has been observed. The relevant option is 'GPUScaling'.'''

        # this fails if it's done for all of them at once, so do it separately
        for i,d in enumerate(displays):
            xtrainfo=''
            curscaling = scaling
            if type(curscaling) == list: curscaling = curscaling[i]
            if curscaling=='default':
                try:
                    sc = nvidia.NVidiaSettings().query(int, 'GPUScaling', None, d)
                    if not sc: continue
                    if (sc&0xffff)==1: curscaling='stretched'
                    elif (sc&0xffff)==2: curscaling='centered'
                    elif (sc&0xffff)==3: curscaling='aspect-scaled'
                    else:
                        self.log.warn('unrecognised scaling value for %s from nvidia-settings: %d'%(d,sc))
                        continue
                    xtrainfo=' (from nvidia-settings configuration)'
                except IOError: continue
            if curscaling=='native':
                self.log.info('setting scaling of display %s to %s%s'%(d, curscaling, xtrainfo))
                self.nv.set_gpu_scaling(self.screen, d, 'native', 'stretched')
            else:
                self.log.info('setting scaling of display %s to %s%s'%(d, curscaling, xtrainfo))
                self.nv.set_gpu_scaling(self.screen, d, 'best-fit', curscaling)

    def get_scaling(self, displays):
        '''return an array of scaling modes for each display'''
        scalings = []
        for d in displays:
            res = self.nv.get_gpu_scaling(self.screen, d)
            if not res: # 'default' on error
                self.log.warning('could not get scaling for screen %s, reverting to "default"'%d)
                scalings.append('default')
            elif res[0]=='native':
                scalings.append('native')
            else:
                scalings.append(res[1])
        return scalings

# vim:ts=4:sw=4:expandtab:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Python-XRandR provides a high level API for the XRandR extension of the
# X.org server. XRandR allows to configure resolution, refresh rate, rotation 
# of the screen and multiple outputs of graphics cards.
#
# Copyright 2007 © Sebastian Heinlein <sebastian.heinlein@web.de>
# Copyright 2007 © Michael Vogt <mvo@ubuntu.com>
# Copyright 2007 © Canonical Ltd.
#
# In many aspects it follows the design of the xrand tool of the X.org, which
# comes with the following copyright:
#
# Copyright © 2001 Keith Packard, member of The XFree86 Project, Inc.
# Copyright © 2002 Hewlett Packard Company, Inc.
# Copyright © 2006 Intel Corporation
#
# And can be downloaded here:
#
# git://anongit.freedesktop.org/git/xorg/app/xrandr
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

__author__ = "Sebastian Heinlein, Michael Vogt"
__version__ = "0.1.1.x"
__status__ = "development"

from ctypes import *
import os


RR_ROTATE_0 = 1
RR_ROTATE_90 = 2
RR_ROTATE_180 = 4
RR_ROTATE_270 = 8
RR_REFLECT_X = 16
RR_REFLECT_Y = 32

RR_CONNECTED = 0
RR_DISCONNECTED = 1
RR_UNKOWN_CONNECTION = 2

RR_BAD_OUTPUT = 0
RR_BAD_CRTC = 1
RR_BAD_MODE = 2

RR_SET_CONFIG_SUCCESS = 0
RR_SET_CONFIG_INVALID_CONFIG_TIME = 1
RR_SET_CONFIG_INVALID_TIME = 2
RR_SET_CONFIG_FAILED = 3

# Flags to keep track of changes
CHANGES_NONE = 0
CHANGES_CRTC = 1
CHANGES_MODE = 2
CHANGES_RELATION = 4
CHANGES_POSITION = 8
CHANGES_ROTATION = 16
CHANGES_REFLECTION = 32
CHANGES_AUTOMATIC = 64
CHANGES_REFRESH = 128
CHANGES_PROPERTY = 256

# Relation information
RELATION_ABOVE = 0
RELATION_BELOW = 1
RELATION_RIGHT_OF = 2
RELATION_LEFT_OF = 3
RELATION_SAME_AS = 4

from core import Screen, xlib, rr

def get_current_display():
    """Returns the currently used display"""
    display_url = os.getenv("DISPLAY")
    dpy = xlib.XOpenDisplay(display_url)
    return dpy

def get_current_screen():
    """Returns the currently used screen"""
    dpy = get_current_display()
    if not dpy: return None
    screen = Screen(dpy)
    return screen

def get_screen_of_display(display, count):
    """Returns the screen of the given display"""
    dpy = xlib.XOpenDisplay(display)
    return Screen(dpy, count)

def get_version():
    """Returns a tuple containing the major and minor version of the xrandr
       extension or None if the extension is not available"""
    major = c_int()
    minor = c_int()
    dpy = get_current_display()
    if not dpy: return None
    res = core.rr.XRRQueryVersion(dpy, byref(major), byref(minor))
    if res:
        return (major.value, minor.value)
    return None

def has_extension():
    """Returns True if the xrandr extension is available"""
    if XRANDR_VERSION:
        return True
    return False

def _check_required_version(version):
    """Raises an exception if the given or a later version of xrandr is not
       available"""
    if XRANDR_VERSION == None or XRANDR_VERSION < version:
        raise UnsupportedRRError(version, XRANDR_VERSION)

XRANDR_VERSION = get_version()

# vim:ts=4:sw=4:et



###############################################################################
# edid.py - EDID parsing routines
#
# code taken from xac-0.6_pre4
# copyright was not clearly indicated, put probably:
#   Copyright 1999-2006 Gentoo Foundation
#   Distributed under the terms of the GNU General Public License v2
#

### FIXME recheck endianness for these structs

from struct import pack, unpack

### EDID Header Magic
EDID_HEADER = "\0\xFF\xFF\xFF\xFF\xFF\xFF\0" 
### ASCII capital letter offset
ASC = 64
SERIAL_DT = "\0\0\0\xFF\0"
ASCII_DT = "\0\0\0\xFE\0"
RANGE_DT = "\0\0\0\xFD\0"
NAME_DT = "\0\0\0\xFC\0"

### Offsets from the edid spec
class Edid:
	def __init__(self, edid=None):
		self.edid = None
		self.valid = 1
		### Parse the passed in edid
		if edid == None:
#			print "Error: Empty EDID"
			self.valid = 0
			edid = "\0" * 128
		### Make sure it a valid edid (header matches)
		if edid[:8] != EDID_HEADER:
#			print "Error: Invalid EDID"
			self.valid = 0
			edid = "\0" * 128
		### Check the checksum 
		sum = 0
		for i in range(128):
			sum = (sum + int(unpack('B', edid[i])[0])) & 0xFF
		if sum != 0:
#			print "Checksum failed, result should be 0, instead it's:", sum
			self.valid = 0
			edid = "\0" * 128

		self.edid = edid

	### ID String
	def get_id_string(self):
		name = unpack('>H',self.edid[8:10])[0]
		### FIXME Be sure to check endianness on an x86 machine
		n = chr(((name >> 10) & 0x1F) + ASC) + chr(((name >> 5) & 0x1F) + ASC) + chr(((name >> 0) & 0x1F) + ASC)
		return n + hex(unpack('>H', self.edid[10:12])[0])[2:].upper()
	
	### Serial Number
	def get_serial_number(self):
		return unpack('>I', self.edid[12:16])[0]

	### Tupple of week, year from EDID
	def get_date(self):
		date = unpack('BB', self.edid[16:18])
		return [date[0], date[1] + 1990]

	### EDID version, revision
	def get_edid_ver(self):
		return unpack('BB', self.edid[18:20])

	### Video input definition struct
	def get_video_input_def(self):
		vid_in = {}
		v = unpack('B', self.edid[20])[0]
		vid_in['digital'] = (v >> 7) & 1
		if vid_in['digital']:
			vid_in['DFP1x'] = (v & 1)	
		else:
			vid_in['video_level'] = (v >> 5) & 3
			vid_in['blank_to_black'] = (v >> 4) & 1
			vid_in['seperate_sync'] = (v >> 3) & 1
			vid_in['composite_sync'] = (v >> 2) & 1
			vid_in['sync_on_green'] = (v >> 1) & 1
			vid_in['serration_vsync'] = (v & 1)
		return vid_in

	### Size of screen (horiz, vert)
	def get_size(self):
		return unpack('BB', self.edid[21:23])

	### Gamma
	def get_gamma(self):
		return (unpack('B', self.edid[23])[0] / 100.0) + 1.0
	
	### Feature Support
	def get_feature_support(self):
		fs = {}
		f = unpack('B', self.edid[24])[0]
		fs['standby'] = (f >> 7) & 1
		fs['suspend'] = (f >> 6) & 1
		fs['active_off'] = (f >> 5) & 1
		fs['display_type'] = (f >> 3) & 3
		fs['rgb'] = (f >> 2) & 1
		fs['prefered_timing'] = (f >> 1) & 1
		fs['gtf'] = (f & 1)
		return fs

	### Color Characteristics
	def get_color_characteristics(self):
		cc = {}
		c = unpack('10B', self.edid[25:35])
		### Check the / 1024 value... (is this right?)
		cc['red_x'] = ((c[2] << 2) + ((c[0] >> 6) & 3)) / 1024.0
		cc['red_y'] = ((c[3] << 2) + ((c[0] >> 4) & 3)) / 1024.0
		cc['green_x'] = ((c[4] << 2) + ((c[0] >> 2) & 3)) / 1024.0
		cc['green_y'] = ((c[5] << 2) + (c[0] & 3)) / 1024.0
		cc['blue_x'] = ((c[6] << 2) + ((c[1] >> 6) & 3)) / 1024.0
		cc['blue_y'] = ((c[7] << 2) + ((c[1] >> 4) & 3)) / 1024.0
		cc['white_x'] = ((c[8] << 2) + ((c[1] >> 2) & 3)) / 1024.0
		cc['white_y'] = ((c[9] << 2) + (c[1] & 3)) / 1024.0

		return cc

	### Established Timings
	def get_timings(self):
		timing = [ [720,400,70], [720,400,88], [640,480,60], [640,480, 67],
			   [640,480,72], [640,480,75], [800,600,56], [800,600, 60],
			   [800,600,72], [800,600,75], [832,624,75], [1024,768, 87],
			   [1024,768,60], [1024,768,70], [1024,768,75], [1280, 1024,75] ]
		my_timing = []

		### Established timings
		est = unpack('BB', self.edid[35:37])
		### This is what the field is set to if not used	
		if est[0] == "\x01":
			est[0] = 0
		elif est[1] == "\x01":
			est[1] = 0	

		for j in range(2):
			for i in range(8):
				if ((est[j] >> i) & 1):
					my_timing.append(timing[i])

		### Add the "standard" timings into the timing array
		std = unpack('>8H', self.edid[38:54])
		man = unpack('B', self.edid[37])[0]

		for i in range(8):
			if ((man >> i) & 1):
				horiz = ((std[i] & 0xFF00) >> 8) * 8 + 248
				aspect = std[i] & 3
				if aspect == 0:
					vert = (horiz * 16) / 10
				elif aspect == 1:
					vert = (horiz * 4) / 3
				elif aspect == 2:
					vert = (horiz * 5) / 4
				else: # aspect == 3
					vert = (horiz * 16) / 9
	
				freq = ((std[i] & 0x00FC) >> 2) + 60
				my_timing.append([horiz, vert, freq])
		return my_timing

	### Monitor Descriptor
	def get_range_dt(self, tag):
		sync = {}
		### If possible, we get the range limit from the range limit tag
		sync['v_min'] = unpack('B',tag[5])[0]
		sync['v_max'] = unpack('B',tag[6])[0]
		sync['h_min'] = unpack('B',tag[7])[0]
		sync['h_max'] = unpack('B',tag[8])[0]
		return sync

	### Detailed Timing (from monitor details)
	def get_timing_dt(self, info):
		### Search for detailed timing info
		timing = {}
		timing['pixel_clock'] = unpack(">H",info[:2])[0]

		timing['horizontal_active'] = unpack("B",info[2])[0] + ((unpack("B",info[4])[0] >> 4) << 8)
		timing['horizontal_blanking'] = unpack("B",info[3])[0] + ((unpack("B",info[4])[0] & 0xF) << 8)

		timing['vertical_active'] = unpack("B",info[5])[0] + (((unpack("B",info[7])[0] >> 4) & 0xF) << 8)
		timing['vertical_blanking'] = unpack("B",info[6])[0] + ((unpack("B",info[7])[0] & 0xF) << 8)

		timing['hsync_offset'] = unpack("B",info[8])[0] + (((unpack("B", info[11])[0] >> 6) & 3) << 8)
		timing['hsync_pulse_width'] = unpack("B",info[9])[0] + (((unpack("B", info[11])[0] >> 4) & 3) << 8)
		timing['vsync_offset'] = (unpack("B",info[10])[0] >> 4) + (((unpack("B", info[11])[0] >> 2) & 3) << 8)
		timing['vsync_pulse_width'] = (unpack("B",info[10])[0] & 0xF) + ((unpack("B", info[11])[0] & 3) << 8)
				
		timing['himage_size'] = unpack("B",info[12])[0] + ((unpack("B",info[14])[0] >> 4) << 8)
		timing['vimage_size'] = unpack("B",info[13])[0] + ((unpack("B",info[14])[0] & 0xF) << 8)
				
		timing['hborder'] = unpack("B",info[15])[0]
		timing['vborder'] = unpack("B",info[16])[0]
				
		timing['interlaced'] = (unpack("B", info[17])[0] >> 7) & 1
		timing['stereo'] = (unpack("B", info[17])[0] >> 5) & 3
		timing['digital_composite'] = (unpack("B", info[17])[0] >> 3) & 3
		timing['variant'] = (unpack("B", info[17])[0] >> 1) & 3
		return timing

	### Parse Monitor Details
	def get_monitor_details(self):
		details = []
		for i in range(4):
			dmt_offset = i * 18 + 54
			tag = self.edid[dmt_offset:dmt_offset+18] 

			if tag[:5] == SERIAL_DT:
				for i in range(13):
					if tag[i + 5] == "\x0A" or tag[i + 5] == "\x00":
						i = i - 1
						break
				size = str(i + 1) + "s"
				details.append(["Serial", unpack(size, tag[5:i + 6])[0]])
			elif tag[:5] == ASCII_DT:
				for i in range(13):
					if tag[i + 5] == "\x0A" or tag[i + 5] == "\x00":
						i = i - 1
						break
				size = str(i + 1) + "s"
				details.append(["ASCII", unpack(size, tag[5:i + 6])[0]])
			elif tag[:5] == RANGE_DT:
				details.append(["Range", self.get_range_dt(tag)])
			elif tag[:5] == NAME_DT:
				for i in range(13):
					if tag[i + 5] == "\x0A" or tag[i + 5] == "\x00":
						i = i - 1
						break
				size = str(i + 1) + "s"
				details.append(["Name", unpack(size, tag[5:i + 6])[0]])
			elif tag[:3] == "\0\0\0" and (0 < unpack('B', tag[3])[0] < 16):
				details.append(["Manufacturer", str(unpack("13B", tag[5:18]))])
			elif tag[:2] != "\0\0":
				details.append(["Detailed Timing", self.get_timing_dt(tag)])
		return details

	### Get a readable name for this monitor
	def get_string_name(self):
		s = ""
		details = self.get_monitor_details()
		for i in details:
			if i[0] == "Name" or i[0] == "ASCII":
				s = s + i[1] + " "
		if not len(s):
			s = self.get_id_string()

		return s.strip()
			

	def has_extension(self):
		return unpack("B", self.edid[126])[0]



###############################################################################
# nvtarget.py - targets for nvidia commands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#        
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License at http://www.gnu.org/licenses/gpl.txt
# By using, editing and/or distributing this software you agree to
# the terms and conditions of this license.

NV_CTRL_TARGET_TYPE_X_SCREEN = 0
NV_CTRL_TARGET_TYPE_GPU = 1
NV_CTRL_TARGET_TYPE_FRAMELOCK = 2
NV_CTRL_TARGET_TYPE_VCSC = 3

###############################################################################
# Targets, to indicate where a command should be executed.
#
class Target:
    _name = None
    _id = None
    _type = None

    def id(self):
        return self._id
    def type(self):
        return self._type
    
    def __str__(self):
        return '<nVidia %s #%d>'%(self._name, self.id())

class GPU(Target):
    def __init__(self, ngpu=0):
        '''Target a GPU'''
        self._id = ngpu
        self._type = NV_CTRL_TARGET_TYPE_GPU
        self._name = 'GPU'

class Screen(Target):
    def __init__(self, nscr=0):
        '''Target an X screen'''
        self._id = nscr
        self._type = NV_CTRL_TARGET_TYPE_X_SCREEN
        self._name = 'X screen'





##############################################################################
# plugin.py - disper plugin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#        
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License at http://www.gnu.org/licenses/gpl.txt
# By using, editing and/or distributing this software you agree to
# the terms and conditions of this license.

import logging

class Plugin:
    def __init__(self, disper):
        self.disper = disper
        self.log = logging.getLogger('disper.plugin.'+self.__class__.__name__.lower())

# vim:ts=4:sw=4:expandtab:



##############################################################################
# metamodes.py - metamode parsing for nvidia GPUs
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#        
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License at http://www.gnu.org/licenses/gpl.txt
# By using, editing and/or distributing this software you agree to
# the terms and conditions of this license.

import re


def res2array(res):
    '''return [w,h] from a resolution that is specified either as a string or
    as an array itself.'''
    if type(res)==str:
        c = res.split('x')
        try:
            if len(c) != 2: raise ValueError
            return map(int, c)
        except ValueError:
            raise Exception('malformed resolution: %s'%res)
    elif type(res)==tuple or type(res)==list:
        return res
    else:
        raise Exception('resolution has wrong type: %s'%str(res))


class MetaModeDisplay:
    '''Display part of a MetaMode'''

    display = None
    physical = None
    virtual = None
    position = None

    def __init__(self, value=None):
        self.set(value)

    def set(self, value):
        self.display = self.physical = self.virtual = self.position = None
        if not value: return
        # split display from its configuration
        r = re.match(r'^\s*(\S*)\s*:\s*(.*?)\s*$', value)
        if not r: raise ValueError('malformed metamode portion: %s'%value)
        self.display, dmm = r.group(1), r.group(2)
        # check display
        if not re.match(r'(DFP|CRT|TV)-\d$', self.display):
            raise Exception("Garbage metamode: restart X before changing the display configuration to avoid crashes in the nVidia driver")
        # parse display configuration
        if dmm == 'NULL': return
        dmmparts = dmm.split()
        self.physical = dmmparts.pop(0)
        try: self.physical = res2array(self.physical)
        except: pass
        if len(dmmparts) > 0 and dmmparts[0][0]=='@':
            self.virtual = dmmparts.pop(0)
            self.virtual = map(int, self.virtual[1:].split('x'))
        if len(dmmparts) > 0:
            r = re.match(r'^([+-]\d+)([+-]\d+)$', dmmparts.pop(0))
            if not r: raise ValueError('malformed metamode portion (bad position): %s'%value)
            self.position = [int(r.group(1)), int(r.group(2))]
        if len(dmmparts) > 0:
            raise ValueError('malformed metamode portion (too many components): %s'%value)

    def __str__(self):
        if not self.display:
            return self.__repr__()
        s = '%s: '%self.display

        if type(self.physical) == list:
            s += '%dx%d'%(self.physical[0],self.physical[1])
        elif not self.physical:
            s += 'NULL'
        else:
            s += str(self.physical)

        if self.virtual:
            s += ' @%dx%d'%(self.virtual[0],self.virtual[1])
        if self.position:
            s += ' %+d%+d'%(self.position[0],self.position[1])
        return s

    def __eq__(self, other):
        # fallback to physical resolution for viewport if not set
        virtualS = self.virtual
        if not virtualS: virtualS = self.physical
        virtualO = other.virtual
        if not virtualO: virtualO = other.physical
        return self.display == other.display and self.physical == other.physical \
            and virtualS == virtualO and self.position == other.position

    def __ne__(self, other):
        return not self.__eq__(other)

    def __nonzero__(self):
        return bool(self.display) and bool(self.physical)



class MetaMode:
    '''A MetaMode line'''

    id = None       # integer id of the modeline, if any
    options = {}    # options for modeline
    metamodes = []  # configuration for each display
    src = None      # source line, required for deleting a metamode

    def __init__(self, value=None):
        self.set(value)

    def set(self, value):
        self.src = value
        self.id = None
        self.options = {}
        self.metamodes = []
        if not value: return
        #opts, sep, line = value.partition('::') #python>=2.5
        opts, line = (value.split('::',1)+['']*2)[:2]
        if not line:
            line = opts
            opts = None
        # parse options
        if opts:
            opts = map(lambda x: x.strip(), opts.split(','))
            for opt in opts:
                #key, sep, val = opt.partition('=') # python>=2.5
                key, val = (opt.split('=',1)+['']*2)[:2]
                self.options[key] = val
            if 'id' in self.options:
                self.id = int(self.options['id'])
        # parse display settings
        for disp in line.split(','):
            mode = MetaModeDisplay(disp)
            if mode.physical:
                self.metamodes.append(mode)

    def __str__(self):
        s = ''
        s += ', '.join(map(lambda x: '%s=%s'%(x,self.options[x]), self.options.keys()))
        s += ' :: '
        s += ', '.join(map(str, self.metamodes))
        return s

    def __eq__(self, other):
        # compare by id if int, str, or MetaMode
        if type(other) == int:
            return self.id == other
        elif type(other) == str:
            return self == MetaMode(other)
        else:
            # Only compare metamodes, options don't matter. Displays that
            # have NULL (for which x.physical isn't defined) don't count.
            displaysS = filter(lambda x: x.physical, self.metamodes)
            displaysO = filter(lambda x: x.physical, other.metamodes)
            if len(displaysS) != len(displaysO): return False
            for m in displaysS:
                if not m in displaysO: return False
            return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __nonzero__(self):
        return len(self.metamodes)>0

    def bounding_size(self):
        '''return the size of the total virtual screen as (w,h)'''
        x,y,w,h = self.bounding_box()
        return w,h

    def bounding_box(self):
        '''return the bounding box coordinates of the total virtual screen
        as (x,y,w,h)'''
        cmin = cmax = None
        for d in self.metamodes:
            size = d.physical
            if d.virtual: size = d.virtual
            dmin = d.position
            dmax = [d.position[0]+size[0], d.position[1]+size[1]]
            if not cmin:
                cmin = dmin
                cmax = dmax
                continue
            cmin = ( min(cmin[0], dmin[0]), min(cmin[1], dmin[1]) )
            cmax = ( max(cmax[0], dmax[0]), max(cmax[1], dmax[1]) )
        return cmin[0], cmin[1], cmax[0]-cmin[0], cmax[1]-cmin[1]


class MetaModeList(list):
    '''A list of MetaModes.'''

    def __init__(self, metamodes=None):
        list.__init__(self)
        if metamodes: 
            for m in metamodes:
                self.append(MetaMode(m))

    def find(self, el):
        '''find a MetaMode by either id or string'''
        for i in self:
            if i == el:
                return i


def metamode_clone(displays, physical, virtual=None):
    '''return a MetaMode that clones the specified displays; physical is the 
    physical resolution specified as [w,h] or "WxH"; virtual is the virtual
    resolution, but can be omitted.'''
    try:
        physical = res2array(physical)
        physical = '%dx%d'%(physical[0],physical[1])
    except: pass
    if virtual:
        virtual = res2array(virtual)
        virtual = ' @%dx%d'%(virtual[0],virtual[1])
    else:
        virtual = ''
    mmline = ', '.join(map(lambda d: '%s: %s%s +0+0'%(d,physical,virtual), displays))
    return MetaMode(mmline)


def metamode_add_extend(metamode, direction, display, physical, virtual=None):
    '''add a display configuration to the supplied metamode; physical and
    virtual (optional) are resolutions specified as [w,h] or "WxH"; direction
    is either 'left'/'right'/'top'/'bottom', and metamode is an existing
    metamode, or None to create a new one.

    Displays are placed along the virtual screen as defined by the displays
    already in metamode. No intelligent packing is performed to minimise
    dead areas. I suggest to pack only in one direction (either horizontal
    or vertical, not both). Dead areas will arise when resolutions differ,
    unless the same virtual resolution is specified for each.

    Note that nvidia-auto-select is not allowed as resolution here.'''
    if physical != 'NULL':
        physical = size = res2array(physical)
        physical = '%dx%d'%(physical[0],physical[1])
    if virtual:
        virtual = size = res2array(virtual)
        virtual = ' @%dx%d'%(virtual[0],virtual[1])
    else:
        virtual = ''
    if not metamode:
        return MetaMode('%s: %s%s +0+0'%(display,physical,virtual))
    x,y,w,h = metamode.bounding_box()
    # negative offsets cannot be done need specification of total virtual
    # resolution which I haven't been able to do using dynamic twinview.
    # so we just keep the origin at 0,0 always and shift displays instead.
    if physical != 'NULL':
        offset = 0,0
        if direction == 'left':
            #offset = x - size[0], 0
            offset = 0, 0
            for d in metamode.metamodes:
                d.position[0] += size[0]
        elif direction == 'right':
            offset = x+w, 0
        elif direction == 'top':
            #offset = 0, y - size[1]
            offset = 0, 0
            for d in metamode.metamodes:
                d.position[1] += size[1]
        elif direction == 'bottom':
            offset = 0, y+h
        else:
            raise Exception('unknown extension direction for metamode: %s'%direction)
        mmdisp = MetaModeDisplay('%s: %s%s %+d%+d'%(display,physical,virtual,offset[0],offset[1]))
    else:
        mmdisp = MetaModeDisplay('%s: NULL'%(display))
    metamode.metamodes.append(mmdisp)
    return metamode


# a little testing
if __name__ == '__main__':
    # this is not a realistic metamode list, but it does contain a lot of
    # combinations that may appear in various configurations (and more).
    metamodesstr = [
        'source=xconfig, id=50, switchable=yes :: CRT-0: 1280x1024 @1280x1024 +1280+0, DFP-0: 1280x1024 @1280x1024 +0+0',
        'source=xconfig, id=51, switchable=yes :: CRT-0: 1024x768 @1024x768 +1024+0, DFP-0: 1024x768 @1024x768 +0+0',
        'source=xconfig, id=60, switchable=no :: CRT-0: 800x600 @800x600 -800-0, DFP-0: NULL',
        'source=xconfig, id=62, switchable=no :: DFP-0: 640x480 @640x480 +640+0, TV-0: 1024x768 @1024x768 +0+0',
        'id=63 :: DFP-9: 640x480 @640x480 +640+0, DFP-2: 123x321 @567x765 +640+0',
        '   id=64    ::  CRT-0  :      800x600    @800x600       +0+0   ',
        'id=65::CRT-1:800x600 @800x600 +0+0',
        'id=66 :: CRT-0: 10x10 @10x10 +1024+0, DFP-0: 10x10 @10x10 +0+0, TV-0: 10x10 @10x10 +0+0',
    ]
    mms = MetaModeList(metamodesstr)

    # make sure all are processed
    if len(mms) != len(metamodesstr):
        print 'ERROR: length %d != %d'%( len(mms), len(metamodestr) )

    # access all of them by id
    for i,mm in enumerate(mms):
        if mms.find(mm.id).src != mm.src:
            print 'ERROR: find by id failed for id %d' % mm.id
        if mms.find(str(mm)).src != mm.src:
            print 'ERROR: find by str failed for id %d' % mm.id
        if MetaMode(metamodesstr[i]) != mm:
            print 'ERROR: find by MetaMode failed for id %d' % mm.id

    # find variations in MetaMode strings
    for mm in [ 'CRT-0: 1024x768 @1024x768 +1024+0, DFP-0: 1024x768 @1024x768 +0+0',
                'CRT-0: 1024x768 +1024+0, DFP-0: 1024x768 +0+0',
                '  CRT-0 :    1024x768   +1024+0   ,DFP-0   :  1024x768     +0+0    ',
                'DFP-0: 1024x768 +0+0, CRT-0: 1024x768 +1024+0']:
        if not mms.find(mm) or mms.find(mm).id != 51:
            print 'ERROR: find variation by str failed: %s' % mm

    for mm in [ 'DFP-9: 640x480 @640x480 +640+0, DFP-2: 123x321 @567x765 +640+0',
                'DFP-2: 123x321 @567x765 +640+0, DFP-9: 640x480 @640x480 +640+0',
                ' DFP-2   :    123x321    @567x765  +640+0  ,  DFP-9 :    640x480    @640x480  +640+0  ']:
        if not mms.find(mm) or mms.find(mm).id != 63:
            print 'ERROR: find variation by str failed: %s' % mm

    # test bounding box sizes
    m = MetaMode('DFP-0: 800x600 +0+0, CRT-0: 800x600 +0+0')
    if m.bounding_size() != (800, 600):
        print 'ERROR: bounding box size: %s'%m
    m = MetaMode('DFP-0: 800x600 @1000x800 +0+0, CRT-0: 800x600 +0+0')
    if m.bounding_size() != (1000, 800):
        print 'ERROR: bounding box size: %s'%m
    m = MetaMode('DFP-0: 800x600 +0+0, CRT-0: 800x600 +900+0')
    if m.bounding_size() != (1700, 600):
        print 'ERROR: bounding box size: %s'%m
    m = MetaMode('DFP-0: 800x600 +0-300, CRT-0: 800x600 +0+0')
    if m.bounding_size() != (800, 900):
        print 'ERROR: bounding box size: %s'%m

    # test metamode creation stuff
    m1 = metamode_clone(['DFP-0','DFP-1','TV-2'], '800x600')
    m2 = MetaMode('DFP-0: 800x600 +0+0, DFP-1: 800x600 +0+0, TV-2: 800x600 +0+0')
    if m1 != m2:
        print 'ERROR: metamode_clone: %s'%str(m2)
    for dir,mmline in [
            ('right', 'CRT-0: 800x600   +0+000, DFP-0: 200x300 +800+000, TV-0: 640x480 +1000+000'),
           #('left',  'CRT-0: 800x600   +0+000, DFP-0: 200x300 -200+000, TV-0: 640x480  -840+000'),
            ('left',  'CRT-0: 800x600 +840+000, DFP-0: 200x300 +640+000, TV-0: 640x480    +0+000'),
           #('top',   'CRT-0: 800x600   +0+000, DFP-0: 200x300   +0-300, TV-0: 640x480    +0-780'),
            ('top',   'CRT-0: 800x600   +0+780, DFP-0: 200x300   +0+480, TV-0: 640x480    +0+000'),
            ('bottom','CRT-0: 800x600   +0+000, DFP-0: 200x300   +0+600, TV-0: 640x480    +0+900') ]:
        m = metamode_add_extend(None, dir, 'CRT-0', [800,600])
        m = metamode_add_extend(m, dir, 'DFP-0', [200,300])
        m = metamode_add_extend(m, dir, 'TV-0', [640,480])
        if m != MetaMode(mmline):
            print 'ERROR: metamode_extend %s: %s'%(dir,str(m))

    print 'tests finished.'


# vim:ts=4:sw=4:expandtab:



##############################################################################
# hook.py - disper script plugin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#        
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License at http://www.gnu.org/licenses/gpl.txt
# By using, editing and/or distributing this software you agree to
# the terms and conditions of this license.

import os
import logging
import subprocess
from plugin import Plugin

class Hook(Plugin):
    '''A hook is a plugin that executes an external command'''

    def __init__(self, disper, script=None):
        Plugin.__init__(self, disper)
        if script: self.set_script(script)
        self._env = os.environ.copy()
        self._env['DISPER_VERSION'] = self.disper.version

    def set_script(self, script):
        '''set the script to execute'''
        self._script = script
        self.log = logging.getLogger('disper.plugin.hook.'+os.path.basename(script))

    def set_layout_clone(self, displays, resolution):
        displays = self._translate_displays(displays)
        self._env['DISPER_DISPLAYS'] = ' '.join(displays)
        self._env['DISPER_LAYOUT'] = 'clone'
        self._env['DISPER_BB_RESOLUTION'] = str(resolution)
        for d in displays:
           self._env['DISPER_RESOLUTION_'+d] = str(resolution)
           # since we don't know it for extend, don't set it here either
           #self._env['DISPER_POSITION_'+d] = '0,0'

    def set_layout_extend(self, displays, layout, resolutions):
        hdisplays = self._translate_displays(displays)
        self._env['DISPER_DISPLAYS'] = ' '.join(hdisplays)
        self._env['DISPER_LAYOUT'] = layout
        bb = [0,0]
        for i in range(len(displays)):
            # TODO move computation to switcher/resolutions.py:ResolutionSelection
            r = resolutions[displays[i]].size()
            if layout in ['right','left']:
                bb[0] += r[0]
                bb[1] = max(bb[1], r[1])
            else:
                bb[1] += r[1]
                bb[0] = max(bb[0], r[0])
            self._env['DISPER_RESOLUTION_'+hdisplays[i]] = str(resolutions[displays[i]])
            # position is backend-dependant, we don't know it here :(
            #self._env['DISPER_POSITION_'+hdisplays[i]] = ''
        self._env['DISPER_BB_RESOLUTION'] = 'x'.join(map(str, bb))

    def call(self, stage):
        '''Call the hook'''
        self._env['DISPER_STAGE'] = stage
        self._env['DISPER_LOG_LEVEL'] = self.log.getEffectiveLevel().__str__()
        cmd = [self._script] + self.disper.argv
        self.log.info('Executing hook: '+' '.join(cmd))
        try: subprocess.Popen(cmd, env=self._env).wait()
        except OSError, e: self.log.warning('Could not execute hook '+self._script+': '+ e.strerror)

    def _translate_displays(self, displays):
        '''replace invalid variable name characters for displays'''
        return map(lambda d: d.replace('-','_'), displays)

# vim:ts=4:sw=4:expandtab:



###############################################################################
# minx.py - an example of a minimal X Protocol interface in python
#
# code taken from NvThermometer by Harry Organs, which was based on the code
# for python-xlib, written by Peter Liljenberg
# http://python-xlib.sourceforge.net/
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#        
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License at http://www.gnu.org/licenses/gpl.txt
# By using, editing and/or distributing this software you agree to
# the terms and conditions of this license.

import struct
import socket
from platform import architecture
import xnet


__XFORMATBYTES = { 'CARD8':1,'CARD16':2,'INT8':1,'INT16':2,
'PAD':1,'BYTE':1,'CARD32':4,'INT32':4 ,'STRING8':1 }

__XFORMATS = { 'CARD8':'B','CARD16':'H','INT8':'b','INT16':'h',
'PAD':'B','BYTE':'B','CARD32':'I','INT32':'i','STRING8':'s0I' }

_ARCHW = architecture()[0]
if _ARCHW.startswith('32bit'): # adjust struct format strings for 32bit
    __XFORMATS['CARD32']='L'
    __XFORMATS['INT32']='l'
    __XFORMATS['STRING8']='s0L'

__XERRORMSG = { 1:'Request error. The major or minor opcode of a request is \
invalid.',
2:'Value error. A request contained a bad argument value.',
3:'Window error. A value for a WINDOW argument does not name a defined \
WINDOW.',
4:'Pixmap error. A value for a PIXMAP argument does not name a defined PIXMAP.',
5:'Atom error. A value for an ATOM argument does not name a defined ATOM.',
6:'Cursor error. A value for a CURSOR argument does not name a defined \
CURSOR.',
7:'Font error. A value for a FONT argument does not name a defined FONT \
or a value for a FONTABLE argument does not name a defined FONT or a \
defined GCONTEXT.',
8:'Match error. InputOnly window used as a DRAWABLE, or GCONTEXT argument \
does not have the same root and depth as the destination DRAWABLE argument, \
or argument(s) fails to match request requirements.',
9:'Drawable error. A value for a DRAWABLE argument does not name a defined \
WINDOW or PIXMAP.',
10:'Access error. Resource access violation or conflict with another client',
11:'Alloc error. The server failed to allocate the requested resource.',
12:'Colormap error. A value for a COLORMAP argument does not name a defined \
COLORMAP.',
13:'GContext error. A value for a GCONTEXT argument does not name a defined \
GCONTEXT.',
14:'IDChoice error. The value chosen for a resource identifier either is not \
included in the range assigned to the client or is already in use.',
15:'Name error. A font or color of the specified name does not exist.',
16:'Length error. The length of a request is shorter or longer than that \
required to minimally contain the arguments, or the length of a request \
exceeds the maximum length accepted by the server.',
17:'Implementation error. The server does not implement some aspect of the \
request.' }

_XServerError__XERRORMSG = __XERRORMSG

class XData:
    '''XData is a simple argument container used to avoid the
    pain and errors of indexing'''

    def __init__( self, f, s, v ):
        self.format = f
        self.size = s
        self.value = v


def encode( *arguments ):
    '''encode takes a variable argument list consisting of XData
    types and returns an encoded byte stream ready to send to the X server.
    the XData args are X type, number of elements, and the value(s). the
    order of fields in the resulting byte stream is determined by the order
    of their respective arguments'''

    bytestream = ''

    for arg in arguments:
        structcode = str(__XFORMATS[arg.format])

        if arg.size == 1:
            bytestream = bytestream + struct.pack( structcode, arg.value )
        else:
            if arg.format.startswith('STRING8'):
                structcode = str(arg.size) + structcode
                bytestream = bytestream + struct.pack( structcode, arg.value )
            else:
                for i in arg.value:
                    bytestream = bytestream + struct.pack( structcode, i )
            
    return bytestream


def decode( binary, *arguments ):
    '''decode takes a byte stream and a variable argument list of XData
    types and returns a dict containing the decoded byte stream.
    the XData args are X type, number of elements, and the key name to
    be associated with the value. the order of arguments determines
    what order the fields will be decoded in'''

    data = binary
    rdict = {}

    for arg in arguments:
        structcode = __XFORMATS[arg.format]
        fsz = __XFORMATBYTES[arg.format]
        
        if not isinstance(arg.size,str):
            asz = arg.size
        else:
            asz = rdict[arg.size]

        sz = asz * fsz
        # workaround for https://bugs.launchpad.net/disper/+bug/908856
        sz = max(sz, struct.calcsize(structcode))

        if asz == 1:
            rdict[arg.value] = struct.unpack( structcode, data[:sz] )[0]
        else:
            structcode = str(asz) + structcode
            sz = struct.calcsize(structcode)
            if arg.format.startswith('STRING8'):
                rdict[arg.value] = struct.unpack( structcode, data[:sz] )[0]
            else:
                rdict[arg.value] = struct.unpack( structcode, data[:sz] )

        data = data[sz:]
            
    return rdict, data


###############################################################################
# Exception class for X server errors
#
class XServerError(Exception):
    '''XServerError is an Exception class to raise X errors.
    this class decodes the error return and selects the message
    to display. the error messages are copied from the X Protocol
    pdf. see the doc for an explanation of the other_info field,
    which contains extra data for some errors.'''

    def __init__(self,encoding):
        xreply, ad = decode( encoding,
        XData('CARD8',1,'Error'),
        XData('CARD8',1,'code'),
        XData('CARD16',1,'sequence_number'),
        XData('CARD32',1,'other_info'),
        XData('CARD16',1,'minor_opcode'),
        XData('CARD8',1,'major_opcode'),
        XData('PAD',21,'unused') )
        
        self.error_code = xreply['code']
        self.sequence_number = xreply['sequence_number']
        self.major_opcode = xreply['major_opcode']
        self.minor_opcode = xreply['minor_opcode']
        self.message = __XERRORMSG[self.error_code]
        self.other_info = xreply['other_info']
        
    def __str__(self):
        return 'X Error ' + str(self.error_code) + ': ' + self.message


        
###############################################################################
# Connection Setup request and replies
#
class XConnectRequest:
    '''XConnectRequest encodes the packet needed to connect to the X server'''

    def __init__(self, byte_order, proto_major,
    proto_minor, auth_name, auth_data ):    
        self.encoding = encode( XData('BYTE',1,byte_order),
        XData('PAD',1,0),
        XData('CARD16',1,proto_major),
        XData('CARD16',1,proto_minor),
        XData('CARD16',1,len(auth_name)),
        XData('CARD16',1,len(auth_data)),
        XData('PAD',2,[0,0]),
        XData('STRING8',len(auth_name),auth_name),  
        XData('STRING8',len(auth_data),auth_data) )

        
class XConnectRefusedReply:
    '''X server reply for failed logon attempt'''

    def __init__(self,encoding):
        xreply, n = decode( encoding,
        XData('BYTE',1,'Failed'),
        XData('BYTE',1,'sz_reason'),
        XData('CARD16',1,'protocol_major_version'),
        XData('CARD16',1,'protocol_minor_version'),
        XData('CARD16',1,'sz_additional'),
        XData('STRING8','sz_reason','reason') )
        
        for n, v in xreply.iteritems():
            setattr( self, n, v )

class XConnectAcceptedReply:
    '''the logon reply. contains all the info needed by
    clients to create windows, etc, as well as various
    server info like vendor name'''

    def __init__(self,encoding):
        xreply, ad = decode( encoding, 
        XData('BYTE',1,'Success'),
        XData('PAD',1,'unused_1'), 
        XData('CARD16',1,'protocol_major_version'),
        XData('CARD16',1,'protocol_minor_version'),
        XData('CARD16',1,'sz_additional'),
        XData('CARD32',1,'release_number'),
        XData('CARD32',1,'resource_id_base'),
        XData('CARD32',1,'resource_id_mask'),
        XData('CARD32',1,'motion_buffer_size'),
        XData('CARD16',1,'sz_vendor'),
        XData('CARD16',1,'maximum_request_length'),
        XData('CARD8',1,'n_SCREENS'),
        XData('BYTE',1,'n_FORMATS'),
        XData('BYTE',1,'image_byte_order'),
        XData('BYTE',1,'bitmap_format_bit_order'),
        XData('CARD8',1,'bitmap_format_scanline_unit'),
        XData('CARD8',1,'bitmap_format_scanline_pad'),
        XData('CARD8',1,'min_keycode'),
        XData('CARD8',1,'max_keycode'),
        XData('PAD',4,'unused_2'),
        XData('STRING8','sz_vendor','vendor') )

        for n, v in xreply.iteritems():
            setattr( self, n, v )
    
        self.pixmap_formats = []
        for p in range(self.n_FORMATS):
            pfe, ad = decode( ad, XData('CARD8',1,'depth'),
            XData('CARD8',1,'bits_per_pixel'),
            XData('CARD8',1,'scanline_pad'),
            XData('PAD',5,'unused') )
            self.pixmap_formats.append(pfe)
        
        self.roots = []
        for s in range(self.n_SCREENS):
            se, ad = decode( ad, XData('CARD32',1,'root'),
            XData('CARD32',1,'default_colormap'),
            XData('CARD32',1,'white_pixel'),
            XData('CARD32',1,'black_pixel'),
            XData('CARD32',1,'current_input-masks'),
            XData('CARD16',1,'width_in_pixels'),
            XData('CARD16',1,'height_in_pixels'),
            XData('CARD16',1,'width_in_millimeters'),
            XData('CARD16',1,'height_in_millimeters'),
            XData('CARD16',1,'min_installed_maps'),
            XData('CARD16',1,'max_installed_maps'),
            XData('CARD32',1,'root_visual'),
            XData('CARD8',1,'backing_stores'),
            XData('CARD8',1,'save_unders'),
            XData('CARD8',1,'root_depth'),
            XData('CARD8',1,'n_allowed_depths') )

            se['allowed_depths'] = []
            for d in range(se['n_allowed_depths']):
                de, ad = decode( ad, XData('CARD8',1,'depth'),
                XData('PAD',1,'unused_1'),
                XData('CARD16',1,'n_VISUALTYPES'),
                XData('PAD',4,'unused_2') )
                
                de['visuals'] = []
                for v in range(de['n_VISUALTYPES']):
                    ve, ad = decode( ad, XData('CARD32',1,'visual_id'),
                    XData('CARD8',1,'class'),
                    XData('CARD8',1,'bits_per_rgb_value'),
                    XData('CARD16',1,'colormap_entries'),
                    XData('CARD32',1,'red_mask'),
                    XData('CARD32',1,'green_mask'),
                    XData('CARD32',1,'blue_mask'),
                    XData('PAD',4,'unused') )

                    de['visuals'].append(ve)
                se['allowed_depths'].append(de)

            self.roots.append(se)



class XConnectAuthenticateReply:
    '''reply sent by secured servers to request client authentication.
    authentication procedures are not defined by the X protcol, so the
    way to handle this one is outside an X protocol interface. Should
    probably raise some kind of exception if unhandled'''

    def __init__(self,encoding):
        xreply, ad = decode( encoding,
        XData('BYTE',1,'Authenticate'),
        XData('PAD',5,'unused'),
        XData('CARD16',1,'sz_additional') )

        for n, v in xreply.iteritems():
            setattr( self, n, v )
        
        rs, ad = decode( ad,
        XData('STRING8',self.xdata['sz_additional']*4,'reason') )
        self.reason = rs['reason']


###############################################################################
# QueryExtension request and reply - opcode 98
#
class XQueryExtensionRequest:
    '''this class wraps the X Protocol Query Extension request. it
    requires the name of the extension to look for as a constructor arg'''

    def __init__(self,exname):
        self.encoding = encode( XData('CARD8',1,98),
        XData('PAD',1,0),
        XData('CARD16',1, 2 + ((len(exname)+(len(exname)%4)) /4) ),
        XData('CARD16',1,len(exname)),
        XData('PAD',2,[0,0]),
        XData('STRING8',len(exname),exname) )

class XQueryExtensionReply:
    '''the reply to a Query Extension request. if attr present is
    0, the extension isn't there. if present is 1, extension exists
    and the extension opcode, base error, and base event are returned'''

    def __init__(self,encoding):
        xreply, ad = decode( encoding,
        XData('CARD8',1,'reply'),
        XData('PAD',1,'unused_1'),
        XData('CARD16',1,'sequence_number'),
        XData('CARD32',1,'reply_length'),
        XData('CARD8',1,'present'),
        XData('CARD8',1,'major_opcode'),
        XData('CARD8',1,'first_event'),
        XData('CARD8',1,'first_error'),
        XData('PAD',20,'unused_2') )

        for n, v in xreply.iteritems():
            setattr( self, n, v )


###############################################################################
# ListExtensions request and reply - opcode 99
#
class XListExtensionsRequest:
    '''this class wraps the X List Extensions request'''

    def __init__(self):
        self.encoding = encode( XData('CARD8',1,99),
        XData('PAD',1,0),
        XData('CARD16',1,1) )

class XListExtensionsReply:
    '''this class wraps the X List Extensions reply. it contains
    the extensions as a list of strings, as well as the number
    of strings in the list and the sequence number of request'''

    def __init__(self,encoding):
        xreply, ad = decode( encoding,
        XData('CARD8',1,'reply'),
        XData('CARD8',1,'n_STRs'),
        XData('CARD16',1,'sequence_number'),
        XData('CARD32',1,'reply_length'),
        XData('PAD',24,'unused') )

        for n, v in xreply.iteritems():
            setattr( self, n, v )

        self.names = []
        for s in range(xreply['n_STRs']):
            sz = struct.unpack( 'B', ad[:1] )[0]
            self.names.append( str(ad[1:sz+1]) )
            ad = ad[sz+1:]



###############################################################################
# Procedures to use the request classes to get info, etc
#
def Xchange( xsock, rq) :
    xreply = ''
    try:
        xsock.send( rq.encoding )
        xreply = xsock.recv(65535) # TODO make sure it fits

    except socket.error, err:
        raise xnet.XConnectionError( 'Network error: %s' % err[1] )

    return xreply


def XConnect():

    name, host, displayno, screenno = xnet.get_X_display()
    xsock = xnet.get_X_socket( host, displayno )
    auth_name, auth_data = xnet.get_X_auth( xsock, name, host, displayno )
    byte_order = xnet.get_X_byteorder()

    rq = XConnectRequest( byte_order, 11, 0, auth_name, auth_data )

    xreply = Xchange( xsock, rq )

    if xreply[0] == '\x00':
        repobj = XConnectRefusedReply(xreply)
        raise xnet.XConnectionError( repobj.reason )

    elif xreply[0] == '\x01':
        repobj = XConnectAcceptedReply(xreply)

    elif xreply[0] == '\x02':
        repobj = XConnectAuthenticateReply(xreply) 
        raise xnet.XConnectionError( repobj.reason )

    else:
        raise xnet.XConnectionError( 'Unknown connection failure' )

    return xsock, repobj


def XListExtensions( xsock ):
    rq = XListExtensionsRequest()
    binrp = Xchange( xsock, rq )

    if binrp[0] == 0:
        raise XServerError( binrp )
    else:
        return XListExtensionsReply( binrp )


def XQueryExtension( xsock, exname ):
    rq = XQueryExtensionRequest(exname)
    binrp = Xchange( xsock, rq )

    if binrp[0] == 0:
        if binrp[1] > 0 and binrp[1] <= 17:
            raise XServerError( binrp )

    return XQueryExtensionReply( binrp )
    





###############################################################################
# nvsettings.py - nvidia-settings configuration file parser
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#        
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License at http://www.gnu.org/licenses/gpl.txt
# By using, editing and/or distributing this software you agree to
# the terms and conditions of this license.

_all = ['NVidiaSettings', 'Screen', 'GPU']

import os
import re
import xnet
import socket
from nvtarget import *

class NVidiaSettings:
    '''nvidia-settings configuration file reader/parser'''

    _opts = None    # list of options: (location, param, device, value)

    def __init__(self, filename=None):
        self.load(filename)

    def load(self, filename=None):
        '''load an nvidia-settings configuration file; use None for the
        default ~/.nvidia-settings-rc'''
        if not filename:
            filename = os.path.expanduser('~/.nvidia-settings-rc')
        self._opts = []
        f = open(filename, 'r')
        for l in f:
            # remove comments
            l = l.split('#',2)[0]
            # remove all spaces (as nvidia-settings does in parse.c)
            l = filter(lambda s: not s.isspace(), l)
            # skip lines without an option assignment
            if not l: continue
            # parse and store
            m = re.match('((.+)\/)?(\w+)(\[(.+)?\])?=(.*)$', l)
            if m:
                # silently ignore malformed lines; TODO should warn
                self._opts.append((m.group(2),m.group(3),m.group(5),m.group(6)))
        f.close()

    def query(self, type, option, target=None, device=None):
        '''retrieve the value from an option for a target. If target is None,
        the first found target is found. Device should only be specified
        for options that require one. If the specified device is not found,
        the value of the option without a device specification is returned,
        if present. This is how nvidia-settings behaves if I'm correct.
        Returns first option found, or None if not found. The option is
        cast to type on success (so that you don't have to check for None
        and then cast it).
        Only options for the X display are found.

        TODO: make more specific targets in configfile get precedence.'''
        # option name
        opts = filter(lambda x: x[1]==option, self._opts)
        # device
        # TODO it's unclear if multiple devices may be specified for a single
        #      option in the configuration file
        if device:
            optsdfl = filter(lambda x: not x[2], opts)
            opts = filter(lambda x: x[2]==device, opts)
            if not opts: opts = optsdfl
        if not target: target = Screen(0)
        # target: only current host for display
        d,host,dno,screen = xnet.get_X_display()
        if not host: host=socket.gethostname()
        host = socket.gethostbyname(host)
        if isinstance(target, Screen): screen = target.id()
        # throw away all options that aren't for us
        for o in list(opts):
            loc = o[0]
            if not loc: continue
            # {host}:{display}.{screen}[target_type:target_id]
            m = re.match('(([\w\.]*):)?(\d+)(\.(\d+))?(\[(\w+):(\d+)\])?$', loc)
            if m:
                chost,cdno,cscreen = m.group(2),m.group(3),m.group(5)
                ctargettype,ctargetval = m.group(7),m.group(8)
                if chost: chost = socket.gethostbyname(chost)
                if (not chost or chost==host) and \
                   (not cdno or int(cdno)==int(dno)) and \
                   (not cscreen or int(cscreen)==int(screen)) and \
                   (not ctargettype or \
                     (ctargettype=='gpu' and isinstance(target, GPU)) or \
                     (ctargettype=='screen' and isinstance(target, Screen))
                   ) and \
                   (not ctargetval or ctargetval==target.id()):
                    # everything matches, keep
                    continue
            else:
                # TODO warn
                pass
            # no match, remove opts from list
            opts.remove(o)
        # return
        if len(opts)>0: return type(opts[0][3])
        else: return None


# some tests
if __name__ == '__main__':
    import os

    # make temporary file with test options
    # Note that each line of the here-document is indented with tabs which are
    # subsequently removed for easy reading. Make sure to use tabs!
    tmpfilename = '/tmp/nvsettings-nvidia-settings-rc-test'
    f = open(tmpfilename, 'w')
    f.write('''
		# test nvidia-settings-rc
	    
		# ConfigProperties:
		RcFileLocale = C
		Foo = Bar
		    
		# Attributes:
		0/SyncToVBlank=0
		1/SyncToVBlank=321
		0/DigitalVibrance=0
		0/DigitalVibrance[DFP-0]=1
		0.0/DigitalVibrance[CRT-0]=2
		localhost:0/TestAttr=hi_there
		127.0.0.1:0.0/OtherAttr=hi_here
        0.0.0.0:0/OtherAttr=not me
        #OtherFoo = BadBar # TODO
		0.0/OtherFoo = OtherBar
		0.1/OtherFoo = MoreBar
		[screen:0]/VideoRam=60
		localhost:0[screen:0]/VideoRam=50
		Plurk=1
		[gpu:0]/Plurk=100
    '''.expandtabs(0))
    f.close()

    nvs = NVidiaSettings(tmpfilename)

    try:
        if nvs.query(str, 'Foo') != 'Bar':
            print 'ERROR: Foo'
        if nvs.query(int, 'SyncToVBlank') != 0:
            print 'ERROR: SyncToVBlank'
        if nvs.query(int, 'SyncToVBlank', Screen(0)) != 0:
            print 'ERROR: 0/SyncToVBlank'
        if nvs.query(int, 'SyncToVBlank', Screen(1)) != 0:
            print 'ERROR: 0.1/SyncToVBlank'
        if nvs.query(int, 'DigitalVibrance') != 0:
            print 'ERROR: DigitalVibrance'
        if nvs.query(int, 'DigitalVibrance', None, 'DFP-0') != 1:
            print 'ERROR: DigitalVibrance[DFP-0]'
        if nvs.query(int, 'DigitalVibrance', None, 'CRT-0') != 2:
            print 'ERROR: DigitalVibrance[CRT-0]'
        if nvs.query(int, 'DigitalVibrance', None, 'CRT-9') != 0:
            print 'ERROR: DigitalVibrance[CRT-9]'

        if nvs.query(str, 'TestAttr') != 'hi_there':
            print 'ERROR: TestAttr'
        if nvs.query(str, 'OtherAttr') != 'hi_here':
            print 'ERROR: OtherAttr'

        # TODO
        #if nvs.query(str, 'OtherFoo') != 'BadBar':
        #    print 'ERROR: OtherFoo'
        if nvs.query(str, 'OtherFoo', Screen(0)) != 'OtherBar':
            print 'ERROR: 0/OtherFoo'
        if nvs.query(str, 'OtherFoo', Screen(1)) != 'MoreBar':
            print 'ERROR: 0.1/OtherFoo'

        os.environ['DISPLAY'] = 'localhost:1'
        if nvs.query(int, 'SyncToVBlank') != 321:
            print 'ERROR: 1/SyncToVBlank'
        os.environ['DISPLAY'] = 'localhost:2'
        if nvs.query(int, 'SyncToVBlank') != None:
            print 'ERROR: 2/SyncToVBlank'

    finally:
        os.unlink(tmpfilename)

# vim:ts=4:sw=4:expandtab:



##############################################################################
# __init__.py - disper plugin system
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#        
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License at http://www.gnu.org/licenses/gpl.txt
# By using, editing and/or distributing this software you agree to
# the terms and conditions of this license.

import os
import logging
from hook import Hook
from plugin import Plugin

class Plugins:

    def __init__(self, disper):
        '''Initialise the plugin system'''
        self.log = logging.getLogger('disper.plugin')
        self.disper = disper
        self._discover()
        self._plugin_names_enabled = []

    def call(self, stage):
        '''Call all plugins that are enabled. Must have called #set_enabled first.'''
        for plugin in self._plugin_names_enabled:
            self.call_plugin(plugin, stage)

    def call_plugin(self, plugin, stage):
        '''Call a plugin by name. Must have called #set_enabled first.'''
        self._plugins[plugin].call(stage)

    def set_layout_clone(self, displays, resolution):
        for plugin in self._plugins:
            self._plugins[plugin].set_layout_clone(displays, resolution)

    def set_layout_extend(self, displays, layout, resolutions):
        for plugin in self._plugins:
            self._plugins[plugin].set_layout_extend(displays, layout, resolutions)

    def set_enabled(self, plugins):
        '''Set the plugins to be executed.
           @param plugins comma-separated list or Python list of plugin names
                  to execute; or 'user' for all user-installed plugins or
                  'all' to include all user and system plugins or 'none' for none.'''
        if not isinstance(plugins, list): plugins = map(lambda x: x.strip(), ','.split(plugins))
        # now expand 'none', 'user' and 'all' plugins
        useplugins = []
        for i in range(len(plugins)):
            if plugins[i] == 'none':
                useplugins = []
            elif plugins[i] == 'user':
                useplugins += self._plugins_user.keys()
            elif plugins[i] == 'all':
                useplugins += self._plugins.keys()
            elif plugins[i] in self._plugins.keys():
                useplugins.append(plugins[i])
            else:
                self.log.warning('Ignoring nonexistant plugin: '+plugins[i])

        # TODO should we only allow each plugin to be called once?

        self.log.info('Enabled plugins: '+' '.join(useplugins))
        self._plugin_names_enabled = useplugins

    def _discover(self):
        '''Create lists of all plugins present'''
        # TODO non-hook plugins
        # find user hooks
        self._plugins_user = {}
        for uhook in self._get_executables(os.path.join(os.getenv('HOME'),'.disper','hooks')):
            name = os.path.splitext(os.path.split(uhook)[1])[0]
            self._plugins_user[name] = Hook(self.disper, uhook)
        self.log.info('Available user hooks: '+', '.join(self._plugins_user))
        # find system hooks
        self._plugins_system = {}
        for uhook in self._get_executables(os.path.join(self.disper.prefix_share, 'hooks')):
            name = os.path.splitext(os.path.split(uhook)[1])[0]
            self._plugins_system[name] = Hook(self.disper, uhook)
        self.log.info('Available system hooks: '+', '.join(self._plugins_user))

        # Now create global list of plugins where user plugins get precedence
        self._plugins = {}
        for i in self._plugins_system: self._plugins[i] = self._plugins_system[i]
        for i in self._plugins_user: self._plugins[i] = self._plugins_user[i]

    def _get_executables(self, directory):
        '''Return full paths of executable files present in directory'''
        hooks = []
        if os.path.isdir(directory):
            for f in os.listdir(directory):
                path = os.path.join(directory, f)
                if os.path.isfile(path): hooks.append(path)
        return hooks


# vim:ts=4:sw=4:expandtab:
